from dataclasses import dataclass
from pathlib import Path
import json
import logging
import os
import re

from tensorflow.python.lib.io import file_io
from transformers import T5Config, T5ForConditionalGeneration
import torch


__all__ = ['CachedT5ModelLoader']
TRANSFO_PREFIX = 'model.ckpt'


@dataclass
class CachedT5ModelLoader:
    url: str
    cache_path: Path
    cache_key: str
    model_type: str = 't5-base'
    flush_cache: bool = False

    def __post_init__(self):
        self.ckpt_url = os.path.join(self.url, 'checkpoint')
        self.model_cache_dir = self.cache_path / self.cache_key
        self.model_cache_dir.mkdir(exist_ok=True, parents=True)
        assert file_io.file_exists(self.ckpt_url), 'checkpoint file missing'
        self.ckpt_prefix = file_io.read_file_to_string(self.ckpt_url)

    def _fix_t5_model(self, model: T5ForConditionalGeneration):
        with torch.no_grad():  # Make more similar to TensorFlow implementation
            model.decoder.block[0].layer[1].EncDecAttention.\
                relative_attention_bias.weight.data.zero_()
        return model

    def load(self) -> T5ForConditionalGeneration:
        try:
            if not self.flush_cache:
                return self._fix_t5_model(
                    T5ForConditionalGeneration.from_pretrained(
                        str(self.model_cache_dir), from_tf=True,
                        force_download=False))
        except (RuntimeError, OSError):
            logging.info('T5 model weights not in cache.')
        m = re.search(r'model_checkpoint_path: "(.+?)"', self.ckpt_prefix)
        assert m is not None, 'checkpoint file malformed'

        # Copy over checkpoint data
        ckpt_patt = re.compile(rf'^{m.group(1)}\.(data-\d+-of-\d+|'
                               rf'index|meta)$')
        for name in file_io.list_directory(self.url):
            if not ckpt_patt.match(name):
                continue
            url = os.path.join(self.url, name)
            url_stat = file_io.stat(url)
            cache_file_path = self.model_cache_dir / ckpt_patt.sub(
                                rf'{TRANSFO_PREFIX}.\1', name)
            try:
                cs = os.stat(str(cache_file_path))
                if cs.st_size == url_stat.length and \
                   cs.st_mtime_ns > url_stat.mtime_nsec and \
                   not self.flush_cache:
                    logging.info(f'Skipping {name}...')
                    continue
            except FileNotFoundError:
                pass
            logging.info(f'Caching {name}...')
            file_io.copy(url, str(cache_file_path), overwrite=True)

        # Transformers expects a model config.json
        config = T5Config.from_pretrained(self.model_type)
        with open(str(self.model_cache_dir / 'config.json'), 'w') as f:
            json.dump(config.__dict__, f, indent=4)
        return self._fix_t5_model(T5ForConditionalGeneration.from_pretrained(
                        str(self.model_cache_dir), from_tf=True,
                        force_download=False))



def convert_to_unicode(text):
    """
    Converts `text` to Unicode (if it's not already) assuming utf-8 input."""
    if isinstance(text, str):
        return text
    elif isinstance(text, bytes):
        return text.decode("utf-8", "ignore")
    else:
        raise ValueError(f"Unsupported string type: f{type(text)}")



from dataclasses import dataclass
from typing import List

from transformers import PreTrainedTokenizer
import torch
import torch.nn as nn

from .tokenize import BatchTokenizer
from pygaggle.rerank.base import TextType


__all__ = ['LongBatchEncoder', 'EncoderOutputBatch', 'SingleEncoderOutput',
           'SpecialTokensCleaner']


@dataclass
class SingleEncoderOutput:
    encoder_output: torch.Tensor
    token_ids: torch.Tensor
    text: TextType


@dataclass
class EncoderOutputBatch:
    encoder_output: List[torch.Tensor]
    token_ids: List[torch.Tensor]
    texts: List[TextType]

    def as_single(self) -> 'SingleEncoderOutput':
        return SingleEncoderOutput(self.encoder_output[0],
                                   self.token_ids[0], self.texts[0])

    def __iter__(self):
        return iter(SingleEncoderOutput(enc_out, token_ids, text) for
                    (enc_out, token_ids, text)
                    in zip(self.encoder_output, self.token_ids, self.texts))


class SpecialTokensCleaner:
    def __init__(self, tokenizer: PreTrainedTokenizer):
        self.special_ids = tokenizer.all_special_ids

    def clean(self, output: SingleEncoderOutput) -> SingleEncoderOutput:
        indices = [idx for idx, tok in enumerate(output.token_ids.tolist())
                   if tok not in self.special_ids]
        return SingleEncoderOutput(output.encoder_output[indices],
                                   output.token_ids[indices], output.text)


class LongBatchEncoder:
    """
    Encodes batches of documents that are longer than the maximum sequence
    length by striding a window across
    the sequence dimension.
    Parameters
    ----------
    encoder : nn.Module
        The encoder module, such as `BertModel`.
    tokenizer : BatchTokenizer
        The batch tokenizer to use.
    max_seq_length : int
        The maximum sequence length, typically 512.
    """
    def __init__(self,
                 encoder: nn.Module,
                 tokenizer: BatchTokenizer,
                 max_seq_length: int = 512):
        self.encoder = encoder
        self.device = next(self.encoder.parameters()).device
        self.tokenizer = tokenizer
        self.msl = max_seq_length

    def encode_single(self, input: TextType) -> SingleEncoderOutput:
        return self.encode([input]).as_single()

    def encode(self, batch_input: List[TextType]) -> EncoderOutputBatch:
        batch_output = []
        batch_ids = []
        for ret in self.tokenizer.traverse(batch_input):
            input_ids = ret.output['input_ids']
            lengths = list(map(len, input_ids))
            batch_ids.extend(map(torch.tensor, input_ids))
            input_ids = [(idx, x) for idx, x in enumerate(input_ids)]
            max_len = min(max(lengths), self.msl)
            encode_lst = [[] for _ in input_ids]
            new_input_ids = [(idx, x[:max_len]) for idx, x in input_ids]
            while new_input_ids:
                attn_mask = [[1] * len(x[1]) +
                             [0] * (max_len - len(x[1]))
                             for x in new_input_ids]
                attn_mask = torch.tensor(attn_mask).to(self.device)
                nonpadded_input_ids = new_input_ids
                new_input_ids = [x + [0] * (max_len - len(x[:max_len]))
                                 for _, x in new_input_ids]
                new_input_ids = torch.tensor(new_input_ids).to(self.device)
                outputs, _ = self.encoder(input_ids=new_input_ids,
                                          attention_mask=attn_mask)
                for (idx, _), output in zip(nonpadded_input_ids, outputs):
                    encode_lst[idx].append(output)

                new_input_ids = [(idx, x[max_len:])
                                 for idx, x in nonpadded_input_ids
                                 if len(x) > max_len]
                max_len = min(max(map(lambda x: len(x[1]), new_input_ids),
                                  default=0), self.msl)

            encode_lst = list(map(torch.cat, encode_lst))
            batch_output.extend(encode_lst)
        return EncoderOutputBatch(batch_output, batch_ids, batch_input)



import spacy
import numpy as np
from pygaggle.rerank.base import Text
from typing import List
from dataclasses import dataclass
from copy import deepcopy


@dataclass
class SegmentGroup:
    """
    'segments' stores the List of document segments
    'doc_end_indexes' stores the index of the last segment of each
        document when 'segment()' converting a 'List[Text]' of documents into
        'List[Text]' of segments. It will be used to split and group segments'
        scores and feed the aggregated scores back to documents in 'aggregate()'
    """
    segments: List[Text]
    doc_end_indexes: List[int]


class SegmentProcessor:
    """
    The SegmentProcessor class is responsible for segmenting documents and aggregating the scores of segments from
    the same document.
    """
    def __init__(self, max_characters=10000000):
        self.nlp = spacy.blank("en")
        self.nlp.add_pipe(self.nlp.create_pipe("sentencizer"))
        self.max_characters = max_characters
        self.aggregate_methods = {
            "max": self._max_aggregate,
            "mean": self._mean_aggregate
        }

    def segment(self, documents: List[Text], seg_size: int, stride: int) -> SegmentGroup:
        """
        Breaks each document into segments.  For example, given a document with sentences [1,2,3,4,5], a seg_size of 3,
        and a stride of 2, the document will be broken into segments [[1, 2, 3], [3, 4, 5], and [5]].  If the document's
        text is empty, a single segment containing the document's title is generated.  Otherwise, the document's title
        is prepended to the document's text.

        :param documents: A list of Text objects, each of which corresponds to an indexed document.
        :param seg_size: The number of sentences each segment should contain.
        :param stride: The number of sentences to advance for the next segment.
        :return: A SegmentGroup containing all the documents' segments and the end index of each document in
        segmented_docs.
        """
        segmented_docs, doc_end_indexes, end_idx = [], [0], 0
        for document in documents:
            doc = self.nlp(document.text[:self.max_characters])
            sentences = [sent.string.strip() for sent in doc.sents]
            # If the text is empty (i.e. there are no sentences), the segment_text is solely the title of the document.
            if len(sentences) == 0:
                segment_text = document.title
                segmented_docs.append(Text(segment_text, dict(docid=document.metadata["docid"])))
                end_idx += 1
                doc_end_indexes.append(int(end_idx))
            else:
                for i in range(0, len(sentences), stride):
                    segment_text = ' '.join(sentences[i:i + seg_size])
                    if document.title and (not document.title == ''):
                        segment_text = document.title + '. ' + segment_text
                    segmented_docs.append(Text(segment_text, dict(docid=document.metadata["docid"])))
                    if i + seg_size >= len(sentences):
                        end_idx += i/stride + 1
                        doc_end_indexes.append(int(end_idx))
                        break
        return SegmentGroup(segmented_docs, doc_end_indexes)

    def aggregate(self, documents: List[Text], segments_group: SegmentGroup, method: str = "max") -> List[Text]:
        """
        Aggregates the scores for each of a document's segments and assigns the aggregated score to the document.
        :param documents: A list of Text objects, each of which corresponds to an indexed document.
        :param segments_group: A SegmentGroup containing all the documents' segments and the end index of each document.
        :param method: The aggregation function to use (default is max).
        :return: The updated list of documents, including scores.
        """
        docs = deepcopy(documents)
        for i in range(len(docs)):
            doc_start_idx = segments_group.doc_end_indexes[i]
            doc_end_idx = segments_group.doc_end_indexes[i+1]
            target_scores = [seg.score for seg in segments_group.segments[doc_start_idx: doc_end_idx]]
            docs[i].score = self.aggregate_methods[method](target_scores)
        return docs

    @staticmethod
    def _max_aggregate(scores):
        return max(scores)

    @staticmethod
    def _mean_aggregate(scores):
        return np.mean(scores)



from pathlib import Path
import os

from pydantic import BaseSettings


class Settings(BaseSettings):
    # Cache settings
    cache_dir: Path = Path(os.getenv('XDG_CACHE_HOME',
                                     str(Path.home() / '.cache'))) / 'covidex'
    flush_cache: bool = False


class MsMarcoSettings(Settings):
    pass


class TRECCovidSettings(Settings):
    pass


class Cord19Settings(Settings):
    # T5 model settings
    t5_model_dir: str = 'gs://neuralresearcher_data/covid/data/model_exp304'
    t5_model_type: str = 't5-base'



from typing import Optional, List
from pathlib import Path
import logging

from pydantic import BaseModel, validator
from transformers import (AutoModel,
                          AutoTokenizer,
                          AutoModelForSequenceClassification,
                          T5ForConditionalGeneration)
import torch

from .args import ArgumentParserBuilder, opt
from pygaggle.rerank.base import Reranker
from pygaggle.rerank.bm25 import Bm25Reranker
from pygaggle.rerank.transformer import (
    UnsupervisedTransformerReranker,
    T5Reranker,
    SequenceClassificationTransformerReranker
)
from pygaggle.rerank.random import RandomReranker
from pygaggle.rerank.similarity import CosineSimilarityMatrixProvider
from pygaggle.model import (SimpleBatchTokenizer,
                            T5BatchTokenizer,
                            RerankerEvaluator,
                            metric_names,
                            TrecWriter)
from pygaggle.data import TRECCovidDataset
from pygaggle.settings import TRECCovidSettings


SETTINGS = TRECCovidSettings()
METHOD_CHOICES = ('transformer', 'bm25', 't5', 'seq_class_transformer',
                  'random')


class DocumentRankingEvaluationOptions(BaseModel):
    task: str
    dataset: Path
    index_dir: Path
    method: str
    model: str
    batch_size: int
    seg_size: int
    seg_stride: int
    aggregate_method: str
    device: str
    from_tf: bool
    metrics: List[str]
    model_type: Optional[str]
    tokenizer_name: Optional[str]

    @validator('task')
    def task_exists(cls, v: str):
        assert v in ['trec-covid']

    @validator('dataset')
    def dataset_exists(cls, v: Path):
        assert v.exists(), 'data directory must exist'
        return v

    @validator('index_dir')
    def index_dir_exists(cls, v: Path):
        assert v.exists(), 'index directory must exist'
        return v

    @validator('model')
    def model_sane(cls, v: str, values, **kwargs):
        method = values['method']
        if method == 'transformer' and v is None:
            raise ValueError('transformer name or path must be specified')
        return v

    @validator('tokenizer_name')
    def tokenizer_sane(cls, v: str, values, **kwargs):
        if v is None:
            return values['model']
        return v


def construct_t5(options: DocumentRankingEvaluationOptions) -> Reranker:
    device = torch.device(options.device)
    model = T5ForConditionalGeneration.from_pretrained(options.model,
                                                       from_tf=options.from_tf).to(device).eval()
    tokenizer = AutoTokenizer.from_pretrained(options.model_type, use_fast=False)
    tokenizer = T5BatchTokenizer(tokenizer, options.batch_size)
    return T5Reranker(model, tokenizer)


def construct_transformer(options:
                          DocumentRankingEvaluationOptions) -> Reranker:
    device = torch.device(options.device)
    model = AutoModel.from_pretrained(options.model,
                                      from_tf=options.from_tf).to(device).eval()
    tokenizer = SimpleBatchTokenizer(AutoTokenizer.from_pretrained(
        options.tokenizer_name, use_fast=False),
        options.batch_size)
    provider = CosineSimilarityMatrixProvider()
    return UnsupervisedTransformerReranker(model, tokenizer, provider)


def construct_seq_class_transformer(options: DocumentRankingEvaluationOptions
                                    ) -> Reranker:
    model = AutoModelForSequenceClassification.from_pretrained(options.model, from_tf=options.from_tf)
    device = torch.device(options.device)
    model = model.to(device).eval()
    tokenizer = AutoTokenizer.from_pretrained(options.tokenizer_name, use_fast=False)
    return SequenceClassificationTransformerReranker(model, tokenizer)


def construct_bm25(options: DocumentRankingEvaluationOptions) -> Reranker:
    return Bm25Reranker(index_path=str(options.index_dir))


def main():
    apb = ArgumentParserBuilder()
    apb.add_opts(opt('--task',
                     type=str,
                     default='trec-covid',
                     help='A string correspondding to the task to execute. By default, this is "trec-covid".'),
                 opt('--dataset',
                     type=Path,
                     required=True,
                     help='Path to the directory containing the topics file, qrels file, and run file.'),
                 opt('--index-dir',
                     type=Path,
                     required=True,
                     help='Path to the input Anserini index.'),
                 opt('--method',
                     required=True,
                     type=str,
                     choices=METHOD_CHOICES,
                     help='Specifies the type of reranker to use.'),
                 opt('--model',
                     required=True,
                     type=str,
                     help='Path to pre-trained model or huggingface model name.'),
                 opt('--output-file',
                     type=Path,
                     default='.',
                     help='A path to the output file.'),
                 opt('--overwrite-output',
                     action='store_true',
                     help='If set to true, output will be overwritten if the output file already exists. Otherwise, '
                          'output will be appended to the existing file.'),
                 opt('--batch-size',
                     '-bsz',
                     type=int,
                     default=96,
                     help='The batch size for tokenization.'),
                 opt('--device',
                     type=str,
                     default='cuda:0',
                     help='The CUDA device to use for reranking.'),
                 opt('--from-tf',
                     action='store_true',
                     help='A boolean of whether the pretrained model is being loaded from a Tensorflow checkpoint. '
                          'If flag is unused, assumed to be false.'),
                 opt('--metrics',
                     type=str,
                     nargs='+',
                     default=metric_names(),
                     choices=metric_names(),
                     help='The list of metrics to collect while evaluating the reranker.'),
                 opt('--model-type',
                     type=str,
                     help='The T5 tokenizer name.'),
                 opt('--tokenizer-name',
                     type=str,
                     help='The name of the tokenizer to pull from huggingface using the AutoTokenizer class. If '
                     'left empty, this will be set to the model name.'),
                 opt('--seg-size',
                     type=int,
                     default=10,
                     help='The number of sentences in each segment. For example, given a document with sentences'
                     '[1,2,3,4,5], a seg_size of 3, and a stride of 2, the document will be broken into segments'
                     '[[1, 2, 3], [3, 4, 5], and [5]].'),
                 opt('--seg-stride',
                     type=int,
                     default=5,
                     help='The number of sentences to increment between each segment. For example, given a document'
                     'with sentences [1,2,3,4,5], a seg_size of 3, and a stride of 2, the document will be broken into'
                     'segments [[1, 2, 3], [3, 4, 5], and [5]].'),
                 opt('--aggregate-method',
                     type=str,
                     default="max",
                     help='Aggregation method for combining scores across sentence segments of the same document.'))
    args = apb.parser.parse_args()
    options = DocumentRankingEvaluationOptions(**vars(args))
    logging.info("Preprocessing Queries & Docs:")
    ds = TRECCovidDataset.from_folder(str(options.dataset))
    examples = ds.to_relevance_examples(str(options.index_dir))
    logging.info("Loading Ranker & Tokenizer:")
    construct_map = dict(transformer=construct_transformer,
                         bm25=construct_bm25,
                         t5=construct_t5,
                         seq_class_transformer=construct_seq_class_transformer,
                         random=lambda _: RandomReranker())
    # Retrieve the correct reranker from the options map based on the input flag.
    reranker = construct_map[options.method](options)
    writer = TrecWriter(args.output_file, args.overwrite_output)
    evaluator = RerankerEvaluator(reranker, options.metrics, writer=writer)
    width = max(map(len, args.metrics)) + 1
    logging.info("Reranking:")
    for metric in evaluator.evaluate_by_segments(examples,
                                                 options.seg_size,
                                                 options.seg_stride,
                                                 options.aggregate_method):
        logging.info(f'{metric.name:<{width}}{metric.value:.5}')


if __name__ == '__main__':
    main()



import abc

import torch

from pygaggle.model.encode import SingleEncoderOutput


__all__ = ['SimilarityMatrixProvider', 'CosineSimilarityMatrixProvider']


class SimilarityMatrixProvider:
    @abc.abstractmethod
    def compute_matrix(self,
                       encoded_query: SingleEncoderOutput,
                       encoded_document: SingleEncoderOutput) -> torch.Tensor:
        pass


class CosineSimilarityMatrixProvider(SimilarityMatrixProvider):
    @torch.no_grad()
    def compute_matrix(self,
                       encoded_query: SingleEncoderOutput,
                       encoded_document: SingleEncoderOutput) -> torch.Tensor:
        query_repr = encoded_query.encoder_output
        doc_repr = encoded_document.encoder_output
        matrix = torch.einsum('mh,nh->mn', query_repr, doc_repr)
        dnorm = doc_repr.norm(p=2, dim=1).unsqueeze(0)
        qnorm = query_repr.norm(p=2, dim=1).unsqueeze(1)
        matrix = (matrix / (qnorm + 1e-7)) / (dnorm + 1e-7)
        return matrix



from typing import List

from .base import Query, Text, Reranker


__all__ = ['IdentityReranker']


class IdentityReranker(Reranker):
    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        return texts



"""
This script creates monoT5 input files for training,
Each line in the monoT5 input file follows the format:
    f'Query: {query} Document: {document} Relevant:\t{label}\n')
"""
import argparse
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument("--triples_train", type=str, required=True,
                    help="tsv file <query>, <positive_document>, <negative_document>")
parser.add_argument("--output_to_t5", type=str, required=True,
                    help="t5 train input file")
args = parser.parse_args()

with open(args.output_to_t5, 'w') as fout_t5:
    for line_num, line in enumerate(tqdm(open(args.triples_train))):
        query, positive_document, negative_document = line.strip().split('\t')
        fout_t5.write(f'Query: {query} Document: {positive_document} Relevant:\ttrue\n')
        fout_t5.write(f'Query: {query} Document: {negative_document} Relevant:\tfalse\n')
print('Done!')



"""
This script convert monoT5 output file to msmarco run file
"""
import collections
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--t5_output", type=str, required=True,
                    help="tsv file with two columns, <label> and <score>")
parser.add_argument("--t5_output_ids", type=str, required=True,
                    help="tsv file with two columns, <query_id> and <doc_id>")
parser.add_argument("--mono_run", type=str, required=True,
                    help="path to output mono run, tsv file, with <query_id>, <doc_id> and <rank>")
args = parser.parse_args()

examples = collections.defaultdict(list)
with open(args.t5_output_ids) as f_gt, open(args.t5_output) as f_pred:
    for line_gt, line_pred in zip(f_gt, f_pred):
        query_id, doc_id = line_gt.strip().split('\t')
        _, score = line_pred.strip().split('\t')
        score = float(score)
        examples[query_id].append((doc_id, score))

with open(args.mono_run, 'w') as fout:
    for query_id, doc_ids_scores in examples.items():
        doc_ids_scores.sort(key=lambda x: x[1], reverse=True)
        for rank, (doc_id, _) in enumerate(doc_ids_scores):
            fout.write(f'{query_id}\t{doc_id}\t{rank + 1}\n')



# flake8: noqa
# There's no way to ignore "F401 '...' imported but unused" warnings in this
# module, but to preserve other warnings. So, don't check this module at all.

from .decode import *
from .encode import *
from .evaluate import *
from .serialize import *
from .tokenize import *
from .writer import *



from dataclasses import dataclass
from functools import lru_cache
import logging
from typing import List, Optional
import json
import re

from pyserini.search import SimpleSearcher

from pygaggle.rerank.base import Query, Text

__all__ = ['RelevanceExample', 'Cord19DocumentLoader', 'Cord19AbstractLoader']


@dataclass
class RelevanceExample:
    query: Query
    documents: List[Text]
    labels: List[bool]


@dataclass
class Cord19Document:
    body_text: str
    ref_entries: str
    abstract: Optional[str] = ''

    @property
    def all_text(self):
        return '\n'.join((self.abstract, self.body_text, self.ref_entries))


@dataclass
class Cord19Abstract:
    title: str
    abstract: str

    def all_text(self):
        return '\n'.join((self.title, self.abstract))


@dataclass
class MsMarcoPassage:
    para_text: str

    @property
    def all_text(self):
        return self.para_text


class Cord19DocumentLoader:
    double_space_pattern = re.compile(r'\s\s+')

    def __init__(self, index_path: str):
        self.searcher = SimpleSearcher(index_path)

    @lru_cache(maxsize=1024)
    def load_document(self, id: str) -> Cord19Document:
        def unfold(entries):
            return '\n'.join(x['text'] for x in entries)
        try:
            article = json.loads(
                self.searcher.doc(id).lucene_document().get('raw'))
        except json.decoder.JSONDecodeError:
            raise ValueError('article not found')
        except AttributeError:
            raise ValueError('document unretrievable')
        ref_entries = article['ref_entries'].values()
        return Cord19Document(unfold(article['body_text']),
                              unfold(ref_entries),
                              abstract=unfold(article['abstract']) if 'abstract' in article else '')


class Cord19AbstractLoader:
    double_space_pattern = re.compile(r'\s\s+')

    def __init__(self, index_path: str):
        self.searcher = SimpleSearcher(index_path)

    @lru_cache(maxsize=1024)
    def load_document(self, id: str) -> Cord19Document:
        try:
            article = json.loads(
                self.searcher.doc(id).lucene_document().get('raw'))
        except json.decoder.JSONDecodeError:
            raise ValueError('article not found')
        except AttributeError as e:
            logging.error(e)
            raise ValueError('document unretrievable')
        return Cord19Abstract(article['csv_metadata']['title'],
                              abstract=article['csv_metadata']['abstract'] if 'abstract' in article else '')


class MsMarcoPassageLoader:
    def __init__(self, index_path: str):
        self.searcher = SimpleSearcher(index_path)

    def load_passage(self, id: str) -> MsMarcoPassage:
        try:
            passage = self.searcher.doc(id).lucene_document().get('raw')
        except AttributeError:
            raise ValueError('passage unretrievable')
        return MsMarcoPassage(passage)



from typing import List, Optional, Mapping, Any
import abc

from pygaggle.rerank.base import Query, Text

__all__ = ['Answer', 'Reader']


class Answer:
    """
    Class representing an answer.
    An answer contains the answer text itself and potentially other metadata.
    Parameters
    ----------
    text : str
        The answer text.
    metadata : Mapping[str, Any]
        Additional metadata and other annotations.
    language: str
        The language of the answer text.
    score : Optional[float]
        The score of the answer.
    ctx_score : Optional[float]
        The context score of the answer.
    total_score : Optional[float]
        The aggregated score of answer score and ctx_score
    """
    def __init__(
        self,
        text: str,
        language: str = "en",
        metadata: Mapping[str, Any] = None,
        score: Optional[float] = 0,
        ctx_score: Optional[float] = 0,
        total_score: Optional[float] = 0,
    ):
        self.text = text
        self.language = language
        if metadata is None:
            metadata = dict()
        self.metadata = metadata
        self.score = score
        self.ctx_score = ctx_score
        self.total_score = total_score

    def aggregate_score(self, weight):
        self.total_score = weight*self.score + (1-weight)*self.ctx_score


class Reader:
    """
    Class representing a Reader.
    A Reader takes a Query and a list of Text and returns a list of Answer.
    """
    @abc.abstractmethod
    def predict(
        self,
        query: Query,
        texts: List[Text],
    ) -> List[Answer]:
        """
        Find answers from a list of texts with respect to a query.
        Parameters
        ----------
        query : Query
            The query.
        texts : List[Text]
            The list of texts.
        Returns
        -------
        List[Answer]
            Predicted list of answers.
        """
        pass



"""Script to convert T5 predictions into a TREC-formatted run."""
import argparse
import collections


parser = argparse.ArgumentParser(
    description='Convert T5 predictions into a TREC-formatted run.')
parser.add_argument('--predictions', type=str, required=True, help='T5 predictions file.')
parser.add_argument('--query_run_ids', type=str, required=True,
                    help='File containing query doc id pairs paired with the T5\'s predictions file.')
parser.add_argument('--output', type=str, required=True, help='run file in the TREC format.')

args = parser.parse_args()

examples = collections.defaultdict(lambda: collections.defaultdict(list))
with open(args.query_run_ids) as f_query_run_ids, open(args.predictions) as f_pred:
    for line_query_doc_id, line_pred in zip(f_query_run_ids, f_pred):
        query_id, doc_id = line_query_doc_id.strip().split('\t')
        _, score = line_pred.strip().split('\t')
        examples[query_id][doc_id].append(float(score))

with open(args.output, 'w') as fout:
    for query_id, doc_ids_scores in examples.items():
        doc_ids_scores = [(doc_id, max(scores)) for doc_id, scores in doc_ids_scores.items()]
        doc_ids_scores.sort(key=lambda x: x[1], reverse=True)
        for rank, (doc_id, score) in enumerate(doc_ids_scores):
            fout.write(f'{query_id} Q0 {doc_id} {rank + 1} {1 / (rank + 1)} T5\n')

print('Done!')



from collections import OrderedDict
from typing import List, Optional, Dict
import abc

from sklearn.metrics import recall_score
from tqdm import tqdm
import numpy as np
import string
import regex as re

from pygaggle.data.kaggle import RelevanceExample
from pygaggle.data.retrieval import RetrievalExample
from pygaggle.rerank.base import Reranker
from pygaggle.reader.base import Reader
from pygaggle.model.writer import Writer
from pygaggle.data.segmentation import SegmentProcessor

__all__ = ['RerankerEvaluator', 'DuoRerankerEvaluator', 'metric_names']
METRIC_MAP = OrderedDict()


class MetricAccumulator:
    name: str = None

    def accumulate(self, scores: List[float], gold: List[RelevanceExample]):
        return

    @abc.abstractmethod
    def value(self):
        return


class MeanAccumulator(MetricAccumulator):
    def __init__(self):
        self.scores = []

    @property
    def value(self):
        return np.mean(self.scores)


class TruncatingMixin:
    def truncated_rels(self, scores: List[float]) -> np.ndarray:
        return np.array(scores)


def register_metric(name):
    def wrap_fn(metric_cls):
        METRIC_MAP[name] = metric_cls
        metric_cls.name = name
        return metric_cls
    return wrap_fn


def metric_names():
    return list(METRIC_MAP.keys())


class TopkMixin(TruncatingMixin):
    top_k: int = None

    def truncated_rels(self, scores: List[float]) -> np.ndarray:
        rel_idxs = sorted(list(enumerate(scores)),
                          key=lambda x: x[1], reverse=True)[self.top_k:]
        scores = np.array(scores)
        scores[[x[0] for x in rel_idxs]] = 0
        return scores


class DynamicThresholdingMixin(TruncatingMixin):
    threshold: float = 0.5

    def truncated_rels(self, scores: List[float]) -> np.ndarray:
        scores = np.array(scores)
        scores[scores < self.threshold * np.max(scores)] = 0
        return scores


class RecallAccumulator(TruncatingMixin, MeanAccumulator):
    def accumulate(self, scores: List[float], gold: RelevanceExample):
        score_rels = self.truncated_rels(scores)
        score_rels[score_rels != 0] = 1
        gold_rels = np.array(gold.labels, dtype=int)
        score = recall_score(gold_rels, score_rels, zero_division=0)
        self.scores.append(score)


class PrecisionAccumulator(TruncatingMixin, MeanAccumulator):
    def accumulate(self, scores: List[float], gold: RelevanceExample):
        score_rels = self.truncated_rels(scores)
        score_rels[score_rels != 0] = 1
        score_rels = score_rels.astype(int)
        gold_rels = np.array(gold.labels, dtype=int)
        sum_score = score_rels.sum()
        if sum_score > 0:
            self.scores.append((score_rels & gold_rels).sum() / sum_score)


@register_metric('precision@1')
class PrecisionAt1Metric(TopkMixin, PrecisionAccumulator):
    top_k = 1


@register_metric('recall@3')
class RecallAt3Metric(TopkMixin, RecallAccumulator):
    top_k = 3


@register_metric('recall@50')
class RecallAt50Metric(TopkMixin, RecallAccumulator):
    top_k = 50


@register_metric('recall@1000')
class RecallAt1000Metric(TopkMixin, RecallAccumulator):
    top_k = 1000


@register_metric('mrr')
class MrrMetric(MeanAccumulator):
    def accumulate(self, scores: List[float], gold: RelevanceExample):
        scores = sorted(list(enumerate(scores)),
                        key=lambda x: x[1], reverse=True)
        rr = next((1 / (rank_idx + 1) for rank_idx, (idx, _) in
                   enumerate(scores) if gold.labels[idx]), 0)
        self.scores.append(rr)


@register_metric('mrr@10')
class MrrAt10Metric(MeanAccumulator):
    def accumulate(self, scores: List[float], gold: RelevanceExample):
        scores = sorted(list(enumerate(scores)), key=lambda x: x[1],
                        reverse=True)
        rr = next((1 / (rank_idx + 1) for rank_idx, (idx, _) in
                   enumerate(scores) if (gold.labels[idx] and rank_idx < 10)),
                  0)
        self.scores.append(rr)


class ThresholdedRecallMetric(DynamicThresholdingMixin, RecallAccumulator):
    threshold = 0.5


class ThresholdedPrecisionMetric(DynamicThresholdingMixin,
                                 PrecisionAccumulator):
    threshold = 0.5


class RerankerEvaluator:
    def __init__(self,
                 reranker: Reranker,
                 metric_names: List[str],
                 use_tqdm: bool = True,
                 writer: Optional[Writer] = None):
        self.reranker = reranker
        self.metrics = [METRIC_MAP[name] for name in metric_names]
        self.use_tqdm = use_tqdm
        self.writer = writer

    def evaluate(self,
                 examples: List[RelevanceExample]) -> List[MetricAccumulator]:
        metrics = [cls() for cls in self.metrics]
        for example in tqdm(examples, disable=not self.use_tqdm):
            scores = [x.score for x in self.reranker.rerank(example.query,
                                                            example.documents)]
            if self.writer is not None:
                self.writer.write(scores, example)
            for metric in metrics:
                metric.accumulate(scores, example)
        return metrics

    def evaluate_by_segments(self,
                             examples: List[RelevanceExample],
                             seg_size: int,
                             stride: int,
                             aggregate_method: str) -> List[MetricAccumulator]:
        metrics = [cls() for cls in self.metrics]
        segment_processor = SegmentProcessor()
        for example in tqdm(examples, disable=not self.use_tqdm):
            segment_group = segment_processor.segment(example.documents, seg_size, stride)
            segment_group.segments = self.reranker.rerank(example.query, segment_group.segments)
            doc_scores = [x.score for x in segment_processor.aggregate(example.documents,
                                                                       segment_group,
                                                                       aggregate_method)]
            if self.writer is not None:
                self.writer.write(doc_scores, example)
            for metric in metrics:
                metric.accumulate(doc_scores, example)
        return metrics


class DuoRerankerEvaluator:
    def __init__(self,
                 mono_reranker: Reranker,
                 duo_reranker: Reranker,
                 metric_names: List[str],
                 mono_hits: int = 50,
                 use_tqdm: bool = True,
                 writer: Optional[Writer] = None):
        self.mono_reranker = mono_reranker
        self.duo_reranker = duo_reranker
        self.mono_hits = mono_hits
        self.metrics = [METRIC_MAP[name] for name in metric_names]
        self.use_tqdm = use_tqdm
        self.writer = writer

    def evaluate(self,
                 examples: List[RelevanceExample]) -> List[MetricAccumulator]:
        metrics = [cls() for cls in self.metrics]
        mono_texts = []
        scores = []
        for ct, example in tqdm(enumerate(examples), total=len(examples), disable=not self.use_tqdm):
            mono_out = self.mono_reranker.rerank(example.query, example.documents)
            mono_texts.append(sorted(enumerate(mono_out), key=lambda x: x[1].score, reverse=True)[:self.mono_hits])
            scores.append(np.array([x.score for x in mono_out]))
        for ct, texts in tqdm(enumerate(mono_texts), total=len(mono_texts), disable=not self.use_tqdm):
            duo_in = list(map(lambda x: x[1], texts))
            duo_scores = [x.score for x in self.duo_reranker.rerank(examples[ct].query, duo_in)]

            scores[ct][list(map(lambda x: x[0], texts))] = duo_scores
            if self.writer is not None:
                self.writer.write(list(scores[ct]), examples[ct])
            for metric in metrics:
                metric.accumulate(list(scores[ct]), examples[ct])
        return metrics

    def evaluate_by_segments(self,
                             examples: List[RelevanceExample],
                             seg_size: int,
                             stride: int,
                             aggregate_method: str) -> List[MetricAccumulator]:
        metrics = [cls() for cls in self.metrics]
        segment_processor = SegmentProcessor()
        for example in tqdm(examples, disable=not self.use_tqdm):
            segment_group = segment_processor.segment(example.documents, seg_size, stride)
            segment_group.segments = self.reranker.rerank(example.query, segment_group.segments)
            doc_scores = [x.score for x in segment_processor.aggregate(example.documents,
                                                                       segment_group,
                                                                       aggregate_method)]
            if self.writer is not None:
                self.writer.write(doc_scores, example)
            for metric in metrics:
                metric.accumulate(doc_scores, example)
        return metrics


class ReaderEvaluator:
    """Class for evaluating a reader.
    Takes in a list of examples (query, texts, ground truth answers),
    predicts a list of answers using the Reader passed in, and
    collects the exact match accuracies between the best answer and
    the ground truth answers given in the example.
    Exact match scoring used is identical to the DPR repository.
    """

    def __init__(
        self,
        reader: Reader,
    ):
        self.reader = reader

    def evaluate(
        self,
        examples: List[RetrievalExample],
        topk_em: List[int] = [50],
        dpr_predictions: Optional[Dict[int, List[Dict[str, str]]]] = None,
    ):
        ems = {k: [] for k in topk_em}
        for example in tqdm(examples):
            answers = self.reader.predict(example.query, example.texts, topk_em)
            ground_truth_answers = example.ground_truth_answers

            topk_prediction = {}

            for k in topk_em:
                best_answer = answers[k][0].text
                em_hit = max([ReaderEvaluator.exact_match_score(best_answer, ga) for ga in ground_truth_answers])
                ems[k].append(em_hit)

                topk_prediction[f'top{k}'] = best_answer

            if dpr_predictions is not None:
                dpr_predictions.append({
                    'question': example.query.text,
                    'prediction': topk_prediction,
                })

        return ems

    @staticmethod
    def exact_match_score(prediction, ground_truth):
        return ReaderEvaluator._normalize_answer(prediction) == ReaderEvaluator._normalize_answer(ground_truth)

    @staticmethod
    def _normalize_answer(s):
        def remove_articles(text):
            return re.sub(r'\b(a|an|the)\b', ' ', text)

        def white_space_fix(text):
            return ' '.join(text.split())

        def remove_punc(text):
            exclude = set(string.punctuation)
            return ''.join(ch for ch in text if ch not in exclude)

        def lower(text):
            return text.lower()

        return white_space_fix(remove_articles(remove_punc(lower(s))))



from typing import List, Optional, Dict
import torch

from transformers import DPRReader, DPRReaderTokenizer

from .base import Reader, Answer
from pygaggle.rerank.base import Query, Text


class DensePassageRetrieverReader(Reader):
    """Class containing the DPR Reader
    Takes in a query and a list of the top passages selected by the retrieval model,
    and predicts a list of the best answer spans from the most relevant passages, reranked.

    Parameters
    ----------
    model : DPR Reader model for predicting start, end, and relevance logits
    tokenizer : DPR Reader tokenizer for encoding input query and texts
    num_spans : Number of answer spans to return
    max_answer_length : Maximum length that an answer span can be
    num_spans_per_passage : Maximum number of answer spans to return per passage
    """

    def __init__(
        self,
        model: DPRReader = None,
        tokenizer: DPRReaderTokenizer = None,
        num_spans: int = 1,
        max_answer_length: int = 10,
        num_spans_per_passage: int = 10,
        batch_size: int = 16,
    ):
        self.model = model or self.get_model()
        self.tokenizer = tokenizer or self.get_tokenizer()
        self.device = next(self.model.parameters(), None).device

        self.num_spans = num_spans
        self.max_answer_length = max_answer_length
        self.num_spans_per_passage = num_spans_per_passage

        self.batch_size = batch_size

    @staticmethod
    def get_model(
        pretrained_model_name_or_path: str = 'facebook/dpr-reader-single-nq-base',
        device: Optional[str] = None,
    ) -> DPRReader:
        device = device or ('cuda' if torch.cuda.is_available() else 'cpu')
        device = torch.device(device)
        return DPRReader.from_pretrained(pretrained_model_name_or_path).to(device).eval()

    @staticmethod
    def get_tokenizer(
        pretrained_tokenizer_name_or_path: str = 'facebook/dpr-reader-single-nq-base',
    ) -> DPRReaderTokenizer:
        return DPRReaderTokenizer.from_pretrained(pretrained_tokenizer_name_or_path)

    def predict(
        self,
        query: Query,
        texts: List[Text],
        milestones: Optional[List[int]] = None,
    ) -> Dict[int, List[Answer]]:
        if milestones is None:
            milestones = [len(texts)]

        answers = {}
        top_answers = []
        prev_milestone = 0

        for milestone in milestones:
            added_texts = texts[prev_milestone: milestone]
            for i in range(0, len(added_texts), self.batch_size):
                encoded_inputs = self.tokenizer(
                    questions=query.text,
                    titles=list(map(lambda t: t.title, added_texts[i: i+self.batch_size])),
                    texts=list(map(lambda t: t.text, added_texts[i: i+self.batch_size])),
                    return_tensors='pt',
                    padding=True,
                    truncation=True,
                    max_length=350,
                )
                input_ids = encoded_inputs['input_ids'].to(self.device)
                attention_mask = encoded_inputs['attention_mask'].to(self.device)
                outputs = self.model(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                )
                outputs.start_logits = outputs.start_logits.cpu().detach().numpy()
                outputs.end_logits = outputs.end_logits.cpu().detach().numpy()
                outputs.relevance_logits = outputs.relevance_logits.cpu().detach().numpy()

                predicted_spans = self.tokenizer.decode_best_spans(
                    encoded_inputs,
                    outputs,
                    self.num_spans,
                    self.max_answer_length,
                    self.num_spans_per_passage,
                )

                for span in predicted_spans:
                    top_answers.append(
                        Answer(
                            text=span.text,
                            score=span.span_score,
                            ctx_score=span.relevance_score,
                        )
                    )

            top_answers = sorted(top_answers, key=lambda x: (-x.ctx_score, -x.score))
            answers[milestone] = top_answers[: self.num_spans]

            prev_milestone = milestone

        return answers



# flake8: noqa
# There's no way to ignore "F401 '...' imported but unused" warnings in this
# module, but to preserve other warnings. So, don't check this module at all.

from .kaggle import *
from .relevance import *
from .msmarco import *
from .trec_covid import *



from typing import Optional, List
from pathlib import Path
import logging

from pydantic import BaseModel, validator
from transformers import (AutoModel,
                          AutoModelForQuestionAnswering,
                          AutoTokenizer,
                          BertForSequenceClassification)
import torch

from .args import ArgumentParserBuilder, opt
from pygaggle.rerank.base import Reranker
from pygaggle.rerank.bm25 import Bm25Reranker
from pygaggle.rerank.transformer import (
    QuestionAnsweringTransformerReranker,
    MonoBERT,
    MonoT5,
    UnsupervisedTransformerReranker
    )
from pygaggle.rerank.random import RandomReranker
from pygaggle.rerank.similarity import CosineSimilarityMatrixProvider
from pygaggle.model import (RerankerEvaluator,
                            SimpleBatchTokenizer,
                            metric_names)
from pygaggle.data import LitReviewDataset
from pygaggle.settings import Cord19Settings


SETTINGS = Cord19Settings()
METHOD_CHOICES = ('transformer', 'bm25', 't5', 'seq_class_transformer',
                  'qa_transformer', 'random')


class KaggleEvaluationOptions(BaseModel):
    dataset: Path
    index_dir: Path
    method: str
    batch_size: int
    device: str
    split: str
    do_lower_case: bool
    metrics: List[str]
    model: Optional[str]
    tokenizer_name: Optional[str]

    @validator('dataset')
    def dataset_exists(cls, v: Path):
        assert v.exists(), 'dataset must exist'
        return v

    @validator('model')
    def model_sane(cls, v: Optional[str], values, **kwargs):
        method = values['method']
        if method == 'transformer' and v is None:
            raise ValueError('transformer name must be specified')
        if v == 'biobert':
            return 'monologg/biobert_v1.1_pubmed'
        return v

    @validator('tokenizer_name')
    def tokenizer_sane(cls, v: str, values, **kwargs):
        if v is None:
            return values['model']
        return v


def construct_t5(options: KaggleEvaluationOptions) -> Reranker:
    model = MonoT5.get_model(options.model,
                             device=options.device)
    tokenizer = MonoT5.get_tokenizer(options.model, batch_size=options.batch_size)
    return MonoT5(model, tokenizer)


def construct_transformer(options: KaggleEvaluationOptions) -> Reranker:
    device = torch.device(options.device)
    try:
        model = AutoModel.from_pretrained(options.model).to(device).eval()
    except OSError:
        model = AutoModel.from_pretrained(options.model,
                                          from_tf=True).to(device).eval()
    tokenizer = SimpleBatchTokenizer(
                    AutoTokenizer.from_pretrained(
                        options.tokenizer_name, do_lower_case=options.do_lower_case),
                    options.batch_size)
    provider = CosineSimilarityMatrixProvider()
    return UnsupervisedTransformerReranker(model, tokenizer, provider)


def construct_seq_class_transformer(options:
                                    KaggleEvaluationOptions) -> Reranker:
    try:
        model = MonoBERT.get_model(options.model, device=options.device)
    except OSError:
        try:
            model = MonoBERT.get_model(
                        options.model,
                        from_tf=True,
                        device=options.device)
        except AttributeError:
            # Hotfix for BioBERT MS MARCO. Refactor.
            BertForSequenceClassification.bias = torch.nn.Parameter(
                                                    torch.zeros(2))
            BertForSequenceClassification.weight = torch.nn.Parameter(
                                                    torch.zeros(2, 768))
            model = BertForSequenceClassification.from_pretrained(
                        options.model, from_tf=True)
            model.classifier.weight = BertForSequenceClassification.weight
            model.classifier.bias = BertForSequenceClassification.bias
            device = torch.device(options.device)
            model = model.to(device).eval()
    tokenizer = MonoBERT.get_tokenizer(
                    options.tokenizer_name, do_lower_case=options.do_lower_case)
    return MonoBERT(model, tokenizer)


def construct_qa_transformer(options: KaggleEvaluationOptions) -> Reranker:
    # We load a sequence classification model first -- again, as a workaround.
    # Refactor
    try:
        model = AutoModelForQuestionAnswering.from_pretrained(
                    options.model)
    except OSError:
        model = AutoModelForQuestionAnswering.from_pretrained(
                    options.model, from_tf=True)
    device = torch.device(options.device)
    model = model.to(device).eval()
    tokenizer = AutoTokenizer.from_pretrained(
                    options.tokenizer_name,
                    do_lower_case=options.do_lower_case,
                    use_fast=False)
    return QuestionAnsweringTransformerReranker(model, tokenizer)


def construct_bm25(options: KaggleEvaluationOptions) -> Reranker:
    return Bm25Reranker(index_path=str(options.index_dir))


def main():
    apb = ArgumentParserBuilder()
    apb.add_opts(opt('--dataset', type=Path, required=True),
                 opt('--index-dir', type=Path, required=True),
                 opt('--method',
                     required=True,
                     type=str,
                     choices=METHOD_CHOICES),
                 opt('--model',
                     type=str,
                     help='Path to pre-trained model or huggingface model name'),
                 opt('--split', type=str, default='nq', choices=('nq', 'kq')),
                 opt('--batch-size', '-bsz', type=int, default=96),
                 opt('--device', type=str, default='cuda:0'),
                 opt('--tokenizer-name', type=str),
                 opt('--do-lower-case', action='store_true'),
                 opt('--metrics',
                     type=str,
                     nargs='+',
                     default=metric_names(),
                     choices=metric_names()),
                 opt('--model-type', type=str))
    args = apb.parser.parse_args()
    options = KaggleEvaluationOptions(**vars(args))
    ds = LitReviewDataset.from_file(str(options.dataset))
    examples = ds.to_senticized_dataset(str(options.index_dir),
                                        split=options.split)
    construct_map = dict(transformer=construct_transformer,
                         bm25=construct_bm25,
                         t5=construct_t5,
                         seq_class_transformer=construct_seq_class_transformer,
                         qa_transformer=construct_qa_transformer,
                         random=lambda _: RandomReranker())
    reranker = construct_map[options.method](options)
    evaluator = RerankerEvaluator(reranker, options.metrics)
    width = max(map(len, args.metrics)) + 1
    stdout = []
    for metric in evaluator.evaluate(examples):
        logging.info(f'{metric.name:<{width}}{metric.value:.5}')
        stdout.append(f'{metric.name}\t{metric.value}')
    print('\n'.join(stdout))


if __name__ == '__main__':
    main()



from collections import OrderedDict, defaultdict
from typing import List
import json
import logging

from pydantic import BaseModel
import scipy.special as sp
import numpy as np

from .relevance import RelevanceExample, Cord19DocumentLoader
from pygaggle.model.tokenize import SpacySenticizer
from pygaggle.rerank.base import Query, Text


__all__ = ['MISSING_ID', 'LitReviewCategory', 'LitReviewAnswer',
           'LitReviewDataset', 'LitReviewSubcategory']


MISSING_ID = '<missing>'


class LitReviewAnswer(BaseModel):
    id: str
    title: str
    exact_answer: str


class LitReviewSubcategory(BaseModel):
    nq_name: str
    kq_name: str
    answers: List[LitReviewAnswer]


class LitReviewCategory(BaseModel):
    name: str
    sub_categories: List[LitReviewSubcategory]


class LitReviewDataset(BaseModel):
    version: str
    categories: List[LitReviewCategory]

    @classmethod
    def from_file(cls, filename: str) -> 'LitReviewDataset':
        with open(filename, encoding='utf-8') as f:
            return cls(**json.load(f))

    def query_answer_pairs(self, split: str = 'nq'):
        return ((subcat.nq_name if split == 'nq' else subcat.kq_name, ans)
                for cat in self.categories
                for subcat in cat.sub_categories
                for ans in subcat.answers)

    def to_senticized_dataset(self,
                              index_path: str,
                              split: str = 'nq') -> List[RelevanceExample]:
        loader = Cord19DocumentLoader(index_path)
        tokenizer = SpacySenticizer()
        example_map = OrderedDict()
        rel_map = OrderedDict()
        for query, document in self.query_answer_pairs(split=split):
            if document.id == MISSING_ID:
                logging.warning(f'Skipping {document.title} (missing ID)')
                continue
            key = (query, document.id)
            try:
                doc = loader.load_document(document.id)
                example_map.setdefault(key, tokenizer(doc.all_text))
            except ValueError as e:
                logging.warning(f'Skipping {document.id} ({e})')
                continue
            sents = example_map[key]
            rel_map.setdefault(key, [False] * len(sents))
            for idx, s in enumerate(sents):
                if document.exact_answer in s:
                    rel_map[key][idx] = True
        mean_stats = defaultdict(list)
        for (_, doc_id), rels in rel_map.items():
            int_rels = np.array(list(map(int, rels)))
            p = int_rels.sum()
            mean_stats['Average spans'].append(p)
            mean_stats['Random P@1'].append(np.mean(int_rels))
            n = len(int_rels) - p
            N = len(int_rels)
            mean_stats['Random R@3'].append(1 - (n * (n - 1) * (n - 2)) / (N * (N - 1) * (N - 2)))
            numer = np.array([sp.comb(n, i) / (N - i) for i in range(0, n + 1)]) * p
            denom = np.array([sp.comb(N, i) for i in range(0, n + 1)])
            rr = 1 / np.arange(1, n + 2)
            rmrr = np.sum(numer * rr / denom)
            mean_stats['Random MRR'].append(rmrr)
            if not any(rels):
                logging.warning(f'{doc_id} has no relevant answers')
        for k, v in mean_stats.items():
            logging.info(f'{k}: {np.mean(v)}')
        return [RelevanceExample(Query(query), list(map(lambda s: Text(s,
                dict(docid=docid)), sents)), rels)
                for ((query, docid), sents), (_, rels) in
                zip(example_map.items(), rel_map.items())]



from collections import Counter
from copy import deepcopy
from typing import List
import math

from pyserini.analysis import get_lucene_analyzer, Analyzer
from pyserini.index import IndexReader
import numpy as np

from .base import Reranker, Query, Text


__all__ = ['Bm25Reranker']


class Bm25Reranker(Reranker):
    def __init__(self,
                 k1: float = 1.6,
                 b: float = 0.75,
                 index_path: str = None):
        self.k1 = k1
        self.b = b
        self.use_corpus_estimator = False
        self.analyzer = Analyzer(get_lucene_analyzer())
        if index_path:
            self.use_corpus_estimator = True
            self.index_utils = IndexReader(index_path)

    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        query_words = self.analyzer.analyze(query.text)
        sentences = list(map(self.analyzer.analyze, (t.text for t in texts)))

        query_words_set = set(query_words)
        sentence_sets = list(map(set, sentences))
        if not self.use_corpus_estimator:
            idfs = {w: math.log(len(sentence_sets) / (1 + sum(int(w in sent)
                    for sent in sentence_sets)))
                    for w in query_words_set}
        mean_len = np.mean(list(map(len, sentences)))
        d_len = len(sentences)

        texts = deepcopy(texts)
        for sent_words, text in zip(sentences, texts):
            tf = Counter(filter(query_words.__contains__, sent_words))
            if self.use_corpus_estimator:
                idfs = {w:
                        self.index_utils.compute_bm25_term_weight(
                                text.metadata['docid'], w) for w in tf}
            score = sum(idfs[w] * tf[w] * (self.k1 + 1) /
                        (tf[w] + self.k1 * (1 - self.b + self.b *
                                            (d_len / mean_len))) for w in tf)
            if np.isnan(score):
                score = 0
            text.score = score
        return texts



"""
This script creates monoT5 input files by taking corpus,
queries and the retrieval run file for the queries and then
create files for monoT5 input. Each line in the monoT5 input
file follows the format:
    f'Query: {query} Document: {document} Relevant:\n')
"""
import argparse
import collections
import re
import spacy
from tqdm import tqdm


def load_corpus(path):
    print('Loading corpus...')
    corpus = {}
    n_text = 0
    with open(path, errors='ignore') as f:
        all_text = ' '.join(f.read().split())
        for raw_text in tqdm(all_text.split('</DOC>')):
            if not raw_text:
                continue
            result = re.search(r'\<DOCNO\>(.*)\<\/DOCNO\>', raw_text)
            if not result:
                continue
            doc_id = result.group(1)
            doc_id = doc_id.strip()

            result = re.search(r'\<TEXT\>(.*)\<\/TEXT\>', raw_text)
            doc_text = ''
            if result:
                doc_text = result.group(1)
                doc_text = doc_text.replace('<P>', ' ').replace('</P>', ' ')
                doc_text = doc_text.strip()
                if doc_text:
                    n_text += 1

            corpus[doc_id] = ' '.join(doc_text.split())

    print(f'Loaded {len(corpus)} docs, {n_text} with texts.')
    return corpus


def load_queries(path):
    """Loads queries into a dict of key: query_id, value: query text."""
    print('Loading queries...')
    queries = {}
    with open(path) as f:
        all_text = ' '.join(f.read().split())

        for query_text in tqdm(all_text.split('</top>')):
            if not query_text:
                continue
            result = re.search(r'\<num\>(.*)\<title\>', query_text)
            query_id = result.group(1)
            query_id = query_id.replace('Number: ', '')
            query_id = ' '.join(query_id.split())

            result = re.search(r'\<title\>(.*)\<desc\>', query_text)
            title = result.group(1)
            title = title.strip()

            result = re.search(r'\<desc\>(.*)\<narr\>', query_text)
            desc = result.group(1)
            desc = desc.replace('Description:', '')
            desc = desc.strip()

            result = re.search(r'\<narr\>(.*)', query_text)
            narr = result.group(1)
            narr = narr.replace('Narrative:', '')
            narr = narr.strip()

            query_text = desc
            if not query_text:
                query_text = title
            queries[query_id] = query_text
    return queries


def load_run(path):
    """Loads run into a dict of key: query_id, value: list of candidate doc
    ids."""
    print('Loading run...')
    run = collections.OrderedDict()
    with open(path) as f:
        for line in tqdm(f):
            query_id, _, doc_title, rank, _, _ = line.split(' ')
            if query_id not in run:
                run[query_id] = []
            run[query_id].append((doc_title, int(rank)))

    # Sort candidate docs by rank.
    sorted_run = collections.OrderedDict()
    for query_id, doc_titles_ranks in run.items():
        sorted(doc_titles_ranks, key=lambda x: x[1])
        doc_titles = [doc_titles for doc_titles, _ in doc_titles_ranks]
        sorted_run[query_id] = doc_titles

    return sorted_run


parser = argparse.ArgumentParser(
    description='Create T5-formatted input from Robust04 queries, run and corpus.')
parser.add_argument('--queries', type=str, required=True,
                    help='Path to TREC-formatted queries (topics) file.')
parser.add_argument('--run', type=str, required=True,
                    help='Path to TREC-formatted run file containing the candidates'
                         ' from the first stage retrieval (e.g., BM25).')
parser.add_argument('--corpus', type=str, required=True,
                    help='path to a single file containing all TREC Disks 4 and 5 documents.')
parser.add_argument('--output_segment_texts', type=str, required=True,
                    help='Path to store segments from Robust04 documents.')
parser.add_argument('--output_segment_query_doc_ids', type=str, required=True,
                    help='Path to store query doc id pairs of each segment.')
parser.add_argument('--max_length', type=int, default=10,
                    help='Maximum number of sentences of each segment.')
parser.add_argument('--stride', type=int, default=5,
                    help='Stride (step) in sentences between each segment.')

args = parser.parse_args()

nlp = spacy.blank("en")
nlp.add_pipe(nlp.create_pipe("sentencizer"))

queries = load_queries(path=args.queries)
run = load_run(path=args.run)
corpus = load_corpus(path=args.corpus)

n_segments = 0
n_docs = 0
n_doc_ids_not_found = 0
print('Converting to segments...')
with open(args.output_segment_texts, 'w') as fout_segment_texts, \
        open(args.output_segment_query_doc_ids,
             'w') as fout_segment_query_doc_ids:

    for query_id, doc_ids in tqdm(run.items(), total=len(run)):
        query_text = queries[query_id]
        for doc_id in doc_ids:
            if doc_id not in corpus:
                n_doc_ids_not_found += 1
                continue
            n_docs += 1
            doc_text = corpus[doc_id]
            doc = nlp(doc_text[:10000])
            sentences = [sent.string.strip() for sent in doc.sents]
            for i in range(0, len(sentences), args.stride):
                segment = ' '.join(sentences[i:i + args.max_length])
                fout_segment_query_doc_ids.write(f'{query_id}\t{doc_id}\n')
                fout_segment_texts.write(
                    f'Query: {query_text} Document: {segment} Relevant:\n')
                n_segments += 1
                if i + args.max_length >= len(sentences):
                    break

print(f'Wrote {n_segments} segments from {n_docs} docs.')
print(f'{n_doc_ids_not_found} doc ids not found in the corpus.')
print('Done!')



import os
import shutil
import tarfile
import unittest
from random import randint
from typing import List
from urllib.request import urlretrieve

from pyserini.search import JSimpleSearcherResult
from pyserini.search import SimpleSearcher

from pygaggle.rerank.base import hits_to_texts, Text, Query, Reranker
from pygaggle.rerank.identity import IdentityReranker


class TestSearch(unittest.TestCase):
    def setUp(self):
        # Download pre-built CACM index
        # Append a random value to avoid filename clashes.
        r = randint(0, 10000000)
        self.collection_url = ('https://github.com/castorini/anserini-data/'
                               'raw/master/CACM/lucene-index.cacm.tar.gz')
        self.tarball_name = 'lucene-index.cacm-{}.tar.gz'.format(r)
        self.index_dir = 'index{}/'.format(r)

        _, _ = urlretrieve(self.collection_url, self.tarball_name)

        tarball = tarfile.open(self.tarball_name)
        tarball.extractall(self.index_dir)
        tarball.close()

        self.searcher = SimpleSearcher(
            f'{self.index_dir}lucene-index.cacm')

    def test_basic(self):
        hits = self.searcher.search('information retrieval')

        self.assertTrue(isinstance(hits, List))

        self.assertTrue(isinstance(hits[0], JSimpleSearcherResult))
        self.assertEqual('CACM-3134', hits[0].docid)
        self.assertEqual(3133, hits[0].lucene_docid)
        self.assertEqual(1500, len(hits[0].contents))
        self.assertEqual(1532, len(hits[0].raw))
        self.assertAlmostEqual(4.76550, hits[0].score, places=5)

        texts = hits_to_texts(hits)
        self.assertEqual(len(hits), len(texts))
        self.assertTrue(isinstance(texts, List))
        self.assertTrue(isinstance(texts[0], Text))

        for i in range(0, len(hits)):
            self.assertEqual(hits[i].raw, texts[i].text)
            self.assertAlmostEqual(hits[i].score, texts[i].score, places=5)

        query = Query('dummy query')
        identity_reranker = IdentityReranker()
        self.assertTrue(isinstance(identity_reranker, Reranker))

        output = identity_reranker.rerank(query, texts)

        # Check that reranked output is indeed the same as the input
        for i in range(0, len(hits)):
            self.assertEqual(texts[i].text, output[i].text)
            self.assertEqual(texts[i].metadata, output[i].metadata)
            self.assertAlmostEqual(texts[i].score, output[i].score, places=5)

        # Check that the identity rerank was not destructive
        texts = []
        for i in range(0, len(hits)):
            self.assertEqual(hits[i].raw, output[i].text)
            self.assertAlmostEqual(hits[i].score, output[i].score, places=5)

    def tearDown(self):
        self.searcher.close()
        os.remove(self.tarball_name)
        shutil.rmtree(self.index_dir)


if __name__ == '__main__':
    unittest.main()



from typing import List, Optional
from pathlib import Path
import logging
import json
import numpy as np
import subprocess

from pydantic import BaseModel

from .args import ArgumentParserBuilder, opt
from pygaggle.reader.base import Reader
from pygaggle.reader.dpr_reader import DensePassageRetrieverReader
from pygaggle.data.retrieval import RetrievalExample
from pygaggle.rerank.base import Query, Text
from pygaggle.model.evaluate import ReaderEvaluator

METHOD_CHOICES = ('dpr')


class PassageReadingEvaluationOptions(BaseModel):
    task: str
    method: str
    retrieval_file: Path
    model_name: Optional[str]
    tokenizer_name: Optional[str]
    num_spans: int
    max_answer_length: int
    num_spans_per_passage: int
    device: str
    batch_size: int
    topk_em: List[int]


def construct_dpr(options: PassageReadingEvaluationOptions) -> Reader:
    model = DensePassageRetrieverReader.get_model(options.model_name, options.device)
    tokenizer = DensePassageRetrieverReader.get_tokenizer(options.tokenizer_name)

    return DensePassageRetrieverReader(model,
                                       tokenizer,
                                       options.num_spans,
                                       options.max_answer_length,
                                       options.num_spans_per_passage,
                                       options.batch_size)


def main():
    apb = ArgumentParserBuilder()
    apb.add_opts(
        opt('--task',
            type=str,
            default='wikipedia'),
        opt('--method',
            type=str,
            required=True,
            choices=METHOD_CHOICES),
        opt('--retrieval-file',
            type=Path,
            required=True,
            help='JSON file containing top passages selected by the retrieval model'),
        opt('--model-name',
            type=str,
            default='facebook/dpr-reader-single-nq-base',
            help='Pretrained model for reader'),
        opt('--tokenizer-name',
            type=str,
            default='facebook/dpr-reader-single-nq-base',
            help='Pretrained model for tokenizer'),
        opt('--num-spans',
            type=int,
            default=1,
            help='Number of answer spans to return'),
        opt('--max-answer-length',
            type=int,
            default=10,
            help='Maximum length that an answer span can be'),
        opt('--num-spans-per-passage',
            type=int,
            default=10,
            help='Maximum number of answer spans to return per passage'),
        opt('--output-file',
            type=Path,
            default=None,
            help='File to output predictions for each example; if not specified, this output will be discarded'),
        opt('--device',
            type=str,
            default='cuda:0',
            help='Device for model computations'),
        opt('--batch-size',
            type=int,
            default=16,
            help='batch size of reader inference'),
        opt('--topk-retrieval',
            type=int,
            default=[],
            nargs='+',
            help='Values of k to print the topk accuracy of the retrieval file'),
        opt('--topk-em',
            type=int,
            default=[50],
            nargs='+',
            help='Values of k to print the topk exact match score'),
    )
    args = apb.parser.parse_args()
    options = PassageReadingEvaluationOptions(**vars(args))

    logging.info("Loading the Retrieval File")
    with open(options.retrieval_file) as f:
        data = json.load(f)

    if args.topk_retrieval:
        logging.info("Evaluating Topk Retrieval Accuracies")
        subprocess.call(['python',
                         'tools/scripts/dpr/evaluate_retrieval.py',
                         '--retrieval',
                         options.retrieval_file,
                         '--topk',
                         *map(str, args.topk_retrieval)])

    logging.info("Loading Reader Model and Tokenizer")
    construct_map = dict(
        dpr=construct_dpr,
    )
    reader = construct_map[options.method](options)

    evaluator = ReaderEvaluator(reader)

    max_topk_passages = max(options.topk_em)
    examples = []
    for _, item in data.items():
        examples.append(
            RetrievalExample(
                query=Query(text=item["question"]),
                texts=list(map(lambda context: Text(text=context["text"].split('\n', 1)[1].replace('""', '"'),
                                                    title=context["text"].split('\n', 1)[0].replace('"', '')),
                               item["contexts"]))[: max_topk_passages],
                ground_truth_answers=item["answers"],
            )
        )
    dpr_predictions = [] if args.output_file is not None else None

    ems = evaluator.evaluate(examples, options.topk_em, dpr_predictions)

    logging.info('Reader completed')

    for k in options.topk_em:
        em = np.mean(np.array(ems[k])) * 100.
        logging.info(f'Top{k}\tExact Match Accuracy: {em}')

    if args.output_file is not None:
        with open(args.output_file, 'w', encoding='utf-8', newline='\n') as f:
            json.dump(dpr_predictions, f, indent=4)


if __name__ == '__main__':
    main()



from collections import defaultdict
from copy import deepcopy
from itertools import permutations
from typing import List

from transformers import (AutoTokenizer,
                          AutoModelForSequenceClassification,
                          PreTrainedModel,
                          PreTrainedTokenizer,
                          T5ForConditionalGeneration)
import torch
from .base import Reranker, Query, Text
from .similarity import SimilarityMatrixProvider
from pygaggle.model import (BatchTokenizer,
                            LongBatchEncoder,
                            QueryDocumentBatch,
                            DuoQueryDocumentBatch,
                            QueryDocumentBatchTokenizer,
                            SpecialTokensCleaner,
                            T5BatchTokenizer,
                            T5DuoBatchTokenizer,
                            greedy_decode)


__all__ = ['MonoT5',
           'DuoT5',
           'UnsupervisedTransformerReranker',
           'MonoBERT',
           'QuestionAnsweringTransformerReranker']


class MonoT5(Reranker):
    def __init__(self,
                 model: T5ForConditionalGeneration = None,
                 tokenizer: QueryDocumentBatchTokenizer = None):
        self.model = model or self.get_model()
        self.tokenizer = tokenizer or self.get_tokenizer()
        self.device = next(self.model.parameters(), None).device

    @staticmethod
    def get_model(pretrained_model_name_or_path: str = 'castorini/monot5-base-msmarco',
                  *args, device: str = None, **kwargs) -> T5ForConditionalGeneration:
        device = device or ('cuda' if torch.cuda.is_available() else 'cpu')
        device = torch.device(device)
        return T5ForConditionalGeneration.from_pretrained(pretrained_model_name_or_path,
                                                          *args, **kwargs).to(device).eval()

    @staticmethod
    def get_tokenizer(pretrained_model_name_or_path: str = 't5-base',
                      *args, batch_size: int = 8, **kwargs) -> T5BatchTokenizer:
        return T5BatchTokenizer(
            AutoTokenizer.from_pretrained(pretrained_model_name_or_path, use_fast=False, *args, **kwargs),
            batch_size=batch_size
        )

    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        texts = deepcopy(texts)
        batch_input = QueryDocumentBatch(query=query, documents=texts)
        for batch in self.tokenizer.traverse_query_document(batch_input):
            input_ids = batch.output['input_ids'].to(self.device)
            attn_mask = batch.output['attention_mask'].to(self.device)
            _, batch_scores = greedy_decode(self.model,
                                            input_ids,
                                            length=1,
                                            attention_mask=attn_mask,
                                            return_last_logits=True)

            # 6136 and 1176 are the indexes of the tokens false and true in T5.
            batch_scores = batch_scores[:, [6136, 1176]]
            batch_scores = torch.nn.functional.log_softmax(batch_scores, dim=1)
            batch_log_probs = batch_scores[:, 1].tolist()
            for doc, score in zip(batch.documents, batch_log_probs):
                doc.score = score
        return texts


class DuoT5(Reranker):
    def __init__(self,
                 model: T5ForConditionalGeneration = None,
                 tokenizer: QueryDocumentBatchTokenizer = None):
        self.model = model or self.get_model()
        self.tokenizer = tokenizer or self.get_tokenizer()
        self.device = next(self.model.parameters(), None).device

    @staticmethod
    def get_model(pretrained_model_name_or_path: str = 'castorini/duot5-base-msmarco',
                  *args, device: str = None, **kwargs) -> T5ForConditionalGeneration:
        device = device or ('cuda' if torch.cuda.is_available() else 'cpu')
        device = torch.device(device)
        return T5ForConditionalGeneration.from_pretrained(pretrained_model_name_or_path,
                                                          *args, **kwargs).to(device).eval()

    @staticmethod
    def get_tokenizer(pretrained_model_name_or_path: str = 't5-base',
                      *args, batch_size: int = 8, **kwargs) -> T5DuoBatchTokenizer:
        return T5DuoBatchTokenizer(
            AutoTokenizer.from_pretrained(pretrained_model_name_or_path, use_fast=False, *args, **kwargs),
            batch_size=batch_size
        )

    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        texts = deepcopy(texts)
        doc_pairs = list(permutations(texts, 2))
        scores = defaultdict(float)
        batch_input = DuoQueryDocumentBatch(query=query, doc_pairs=doc_pairs)
        for batch in self.tokenizer.traverse_duo_query_document(batch_input):
            input_ids = batch.output['input_ids'].to(self.device)
            attn_mask = batch.output['attention_mask'].to(self.device)
            _, batch_scores = greedy_decode(self.model,
                                            input_ids,
                                            length=1,
                                            attention_mask=attn_mask,
                                            return_last_logits=True)

            # 6136 and 1176 are the indexes of the tokens false and true in T5.
            batch_scores = batch_scores[:, [6136, 1176]]
            batch_scores = torch.nn.functional.softmax(batch_scores, dim=1)
            batch_probs = batch_scores[:, 1].tolist()
            for doc, score in zip(batch.doc_pairs, batch_probs):
                scores[doc[0].metadata['docid']] += score
                scores[doc[1].metadata['docid']] += (1 - score)

        for text in texts:
            text.score = scores[text.metadata['docid']]
        return texts


class UnsupervisedTransformerReranker(Reranker):
    methods = dict(max=lambda x: x.max().item(),
                   mean=lambda x: x.mean().item(),
                   absmean=lambda x: x.abs().mean().item(),
                   absmax=lambda x: x.abs().max().item())

    def __init__(self,
                 model: PreTrainedModel,
                 tokenizer: BatchTokenizer,
                 sim_matrix_provider: SimilarityMatrixProvider,
                 method: str = 'max',
                 clean_special: bool = True,
                 argmax_only: bool = False):
        assert method in self.methods, 'inappropriate scoring method'
        self.model = model
        self.tokenizer = tokenizer
        self.encoder = LongBatchEncoder(model, tokenizer)
        self.sim_matrix_provider = sim_matrix_provider
        self.method = method
        self.clean_special = clean_special
        self.cleaner = SpecialTokensCleaner(tokenizer.tokenizer)
        self.device = next(self.model.parameters(), None).device
        self.argmax_only = argmax_only

    @torch.no_grad()
    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        encoded_query = self.encoder.encode_single(query)
        encoded_documents = self.encoder.encode(texts)
        texts = deepcopy(texts)
        max_score = None
        for enc_doc, text in zip(encoded_documents, texts):
            if self.clean_special:
                enc_doc = self.cleaner.clean(enc_doc)
            matrix = self.sim_matrix_provider.compute_matrix(encoded_query,
                                                             enc_doc)
            score = self.methods[self.method](matrix) if matrix.size(1) > 0 \
                else -10000
            text.score = score
            max_score = score if max_score is None else max(max_score, score)
        if self.argmax_only:
            for text in texts:
                if text.score != max_score:
                    text.score = max_score - 10000
        return texts


class MonoBERT(Reranker):
    def __init__(self,
                 model: PreTrainedModel = None,
                 tokenizer: PreTrainedTokenizer = None):
        self.model = model or self.get_model()
        self.tokenizer = tokenizer or self.get_tokenizer()
        self.device = next(self.model.parameters(), None).device

    @staticmethod
    def get_model(pretrained_model_name_or_path: str = 'castorini/monobert-large-msmarco',
                  *args, device: str = None, **kwargs) -> AutoModelForSequenceClassification:
        device = device or ('cuda' if torch.cuda.is_available() else 'cpu')
        device = torch.device(device)
        return AutoModelForSequenceClassification.from_pretrained(pretrained_model_name_or_path,
                                                                  *args, **kwargs).to(device).eval()

    @staticmethod
    def get_tokenizer(pretrained_model_name_or_path: str = 'bert-large-uncased',
                      *args, **kwargs) -> AutoTokenizer:
        return AutoTokenizer.from_pretrained(pretrained_model_name_or_path, use_fast=False, *args, **kwargs)

    @torch.no_grad()
    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        texts = deepcopy(texts)
        for text in texts:
            ret = self.tokenizer.encode_plus(query.text,
                                             text.text,
                                             max_length=512,
                                             truncation=True,
                                             return_token_type_ids=True,
                                             return_tensors='pt')
            input_ids = ret['input_ids'].to(self.device)
            tt_ids = ret['token_type_ids'].to(self.device)
            output, = self.model(input_ids, token_type_ids=tt_ids, return_dict=False)
            if output.size(1) > 1:
                text.score = torch.nn.functional.log_softmax(
                                output, 1)[0, -1].item()
            else:
                text.score = output.item()
        return texts


class QuestionAnsweringTransformerReranker(Reranker):
    def __init__(self, model: PreTrainedModel, tokenizer: PreTrainedTokenizer):
        self.tokenizer = tokenizer
        self.model = model
        self.device = next(model.parameters()).device

    @torch.no_grad()
    def rerank(self, query: Query, texts: List[Text]) -> List[Text]:
        texts = deepcopy(texts)
        for text in texts:
            ret = self.tokenizer.encode_plus(query.text,
                                             text.text,
                                             max_length=512,
                                             truncation=True,
                                             return_tensors='pt',
                                             return_token_type_ids=True)
            input_ids = ret['input_ids'].to(self.device)
            tt_ids = ret['token_type_ids'].to(self.device)
            start_scores, end_scores = self.model(input_ids,
                                                  token_type_ids=tt_ids,
                                                  return_dict=False)
            start_scores = start_scores[0]
            end_scores = end_scores[0]
            start_scores[(1 - tt_ids[0]).bool()] = -5000
            end_scores[(1 - tt_ids[0]).bool()] = -5000
            smax_val, smax_idx = start_scores.max(0)
            emax_val, emax_idx = end_scores.max(0)
            text.score = max(smax_val.item(), emax_val.item())
        return texts



"""
This script convert duoT5 output file to msmarco run file
"""
import argparse
import collections
import numpy as np
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument("--t5_output", type=str, required=True,
                    help="tsv file with two columns, <label> and <score>")
parser.add_argument("--t5_output_ids", type=str, required=True,
                    help="tsv file with five columns <query_id>, <doc_id_a>, <doc_id_b>, <rank_a> and <rank_b>")
parser.add_argument("--input_run", type=str, required=True,
                    help="path to input run, tsv file, with <query_id>, <doc_id> and <rank>")
parser.add_argument("--duo_run", type=str, required=True,
                    help="path to output duo run, tsv file, with <query_id>, <doc_id> and <rank>")
parser.add_argument("--top_k", type=int, default=50,
                    help="top-k pointwise hits to be reranked by pairwise ranker")
parser.add_argument("--aggregate", type=str, default="sym_sum",
                    help="aggregation technique: one of sum, sym_sum, log_sum or sym_log_sum")

args = parser.parse_args()


def load_run(path):
    """Loads run into a dict of key: query_id, value: list of candidate doc
    ids."""

    # We want to preserve the order of runs so we can pair the run file with
    # the TFRecord file.
    print('Loading run...')
    run = collections.OrderedDict()
    with open(path) as f:
        for line in tqdm(f):
            query_id, doc_title, rank = line.split('\t')
            if query_id not in run:
                run[query_id] = []
            run[query_id].append((doc_title, int(rank)))

    # Sort candidate docs by rank.
    print('Sorting candidate docs by rank...')
    sorted_run = collections.OrderedDict()
    for query_id, doc_titles_ranks in tqdm(run.items()):
        sorted(doc_titles_ranks, key=lambda x: x[1])
        doc_titles = [doc_titles for doc_titles, _ in doc_titles_ranks]
        sorted_run[query_id] = doc_titles

    return sorted_run


input_run = load_run(path=args.input_run)
examples = collections.defaultdict(dict)
with open(args.t5_output_ids) as f_gt, open(args.t5_output) as f_pred:
    for line_gt, line_pred in zip(f_gt, f_pred):
        query_id, doc_id_a, doc_id_b, ct_a, ct_b = line_gt.strip().split('\t')
        _, score = line_pred.strip().split('\t')
        score = float(score)
        if int(ct_a) < args.top_k and int(ct_b) < args.top_k:
            if doc_id_a not in examples[query_id]:
                examples[query_id][doc_id_a] = 0
            if "log" not in args.aggregate:
                score = np.exp(score)
            examples[query_id][doc_id_a] += score
            if "sym" in args.aggregate:
                if doc_id_b not in examples[query_id]:
                    examples[query_id][doc_id_b] = 0
                if "log" in args.aggregate:
                    score_b = np.log(1 - np.exp(score))
                else:
                    score_b = 1 - score
                examples[query_id][doc_id_b] += score_b

for qid in examples:
    examples[qid] = list(examples[qid].items())

with open(args.duo_run, 'w') as fout:
    for query_id, doc_ids_scores in examples.items():
        doc_ids_scores.sort(key=lambda x: x[1], reverse=True)
        for rank, (doc_id, _) in enumerate(doc_ids_scores):
            fout.write(f'{query_id}\t{doc_id}\t{rank + 1}\n')
        input_offset = len(doc_ids_scores)
        for rank, doc_id in enumerate(input_run[query_id]):
            if rank < input_offset:
                continue
            fout.write(f'{query_id}\t{doc_id}\t{rank + 1}\n')



"""
This script creates monoT5 input files by taking corpus,
queries and the retrieval run file for the queries and then
create files for monoT5 input. Each line in the monoT5 input
file follows the format:
    f'Query: {query} Document: {document} Relevant:\n')
"""
import collections
from tqdm import tqdm
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--queries", type=str, required=True,
                    help="tsv file with two columns, <query_id> and <query_text>")
parser.add_argument("--run", type=str, required=True,
                    help="tsv file with three columns <query_id>, <doc_id> and <rank>")
parser.add_argument("--corpus", type=str, required=True)
parser.add_argument("--t5_input", type=str, required=True,
                    help="path to store t5_input, txt format")
parser.add_argument("--t5_input_ids", type=str, required=True,
                    help="path to store the query-doc ids of t5_input, tsv format")
args = parser.parse_args()


def load_corpus(path):
    print('Loading corpus...')
    corpus = {}
    with open(path) as f:
        for line in tqdm(f):
            doc_id, doc = line.rstrip().split('\t')
            corpus[doc_id] = doc
    return corpus


def load_queries(path):
    """Loads queries into a dict of key: query_id, value: query text."""
    print('Loading queries...')
    queries = {}
    with open(path) as f:
        for line in tqdm(f):
            query_id, query = line.rstrip().split('\t')
            queries[query_id] = query
    return queries


def load_run(path):
    """Loads run into a dict of key: query_id, value: list of candidate doc
    ids."""

    # We want to preserve the order of runs so we can pair the run file with
    # the TFRecord file.
    print('Loading run...')
    run = collections.OrderedDict()
    with open(path) as f:
        for line in tqdm(f):
            query_id, doc_title, rank = line.split('\t')
            if query_id not in run:
                run[query_id] = []
            run[query_id].append((doc_title, int(rank)))

    # Sort candidate docs by rank.
    print('Sorting candidate docs by rank...')
    sorted_run = collections.OrderedDict()
    for query_id, doc_titles_ranks in tqdm(run.items()):
        sorted(doc_titles_ranks, key=lambda x: x[1])
        doc_titles = [doc_titles for doc_titles, _ in doc_titles_ranks]
        sorted_run[query_id] = doc_titles

    return sorted_run


corpus = load_corpus(path=args.corpus)
queries = load_queries(path=args.queries)
run = load_run(path=args.run)

print("Writing t5 input and ids")
with open(args.t5_input, 'w') as fout_t5, open(args.t5_input_ids, 'w') as fout_tsv:
    for num_examples, (query_id, candidate_doc_ids) in enumerate(
            tqdm(run.items(), total=len(run))):
        query = queries[query_id]
        for candidate_doc_id in candidate_doc_ids:
            fout_t5.write(
                f'Query: {query} Document: {corpus[candidate_doc_id]} Relevant:\n')
            fout_tsv.write(f'{query_id}\t{candidate_doc_id}\n')



from dataclasses import dataclass
from typing import List

from pygaggle.rerank.base import Query, Text


@dataclass
class RetrievalExample:
    query: Query
    texts: List[Text]
    ground_truth_answers: List[List[str]]



from typing import Union, Tuple

from transformers import PreTrainedModel
import torch

__all__ = ['greedy_decode']

DecodedOutput = Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]


@torch.no_grad()
def greedy_decode(model: PreTrainedModel,
                  input_ids: torch.Tensor,
                  length: int,
                  attention_mask: torch.Tensor = None,
                  return_last_logits: bool = True) -> DecodedOutput:
    decode_ids = torch.full((input_ids.size(0), 1),
                            model.config.decoder_start_token_id,
                            dtype=torch.long).to(input_ids.device)
    encoder_outputs = model.get_encoder()(input_ids, attention_mask=attention_mask)
    next_token_logits = None
    for _ in range(length):
        model_inputs = model.prepare_inputs_for_generation(
            decode_ids,
            encoder_outputs=encoder_outputs,
            past=None,
            attention_mask=attention_mask,
            use_cache=True)
        outputs = model(**model_inputs)  # (batch_size, cur_len, vocab_size)
        next_token_logits = outputs[0][:, -1, :]  # (batch_size, vocab_size)
        decode_ids = torch.cat([decode_ids,
                                next_token_logits.max(1)[1].unsqueeze(-1)],
                               dim=-1)
    if return_last_logits:
        return decode_ids, next_token_logits
    return decode_ids



import os
from collections import OrderedDict, defaultdict
from typing import List, Set, DefaultDict
import logging
from itertools import permutations

from pydantic import BaseModel
import scipy.special as sp
import numpy as np
from tqdm import tqdm

from .relevance import RelevanceExample, Cord19AbstractLoader
from pygaggle.rerank.base import Query, Text
from pygaggle.data.unicode import convert_to_unicode
import xml.etree.ElementTree as ElementTree


__all__ = ['TRECCovidExample', 'TRECCovidDataset']


class TRECCovidExample(BaseModel):
    qid: str
    text: str
    candidates: List[str]
    relevant_candidates: Set[str]


class TRECCovidDataset(BaseModel):
    examples: List[TRECCovidExample]

    @classmethod
    def load_qrels(cls, path: str) -> DefaultDict[str, Set[str]]:
        qrels = defaultdict(set)
        with open(path) as f:
            for _, line in enumerate(f):
                qid, _, doc_id, relevance = filter(None, line.rstrip().split(' '))
                if int(relevance) >= 1:
                    qrels[qid].add(doc_id)
        return qrels

    @classmethod
    def load_run(cls, path: str):
        '''Returns OrderedDict[str, List[str]]'''
        run = OrderedDict()
        with open(path) as f:
            for _, line in enumerate(f):
                # Line is of the format {qid}, QO, {docid}, {rank}, {score},
                # {tag}.
                qid, _, doc_title, rank, score, _ = line.split(' ')
                if qid not in run:
                    run[qid] = []
                run[qid].append((doc_title, int(rank)))
        sorted_run = OrderedDict()
        for qid, doc_titles_ranks in run.items():
            sorted(doc_titles_ranks, key=lambda x: x[1])
            doc_titles = [doc_titles for doc_titles, _ in doc_titles_ranks]
            sorted_run[qid] = doc_titles
        return sorted_run

    # Should this be queries or topics?
    @classmethod
    def load_queries(cls,
                     path: str,
                     qrels: DefaultDict[str, Set[str]],
                     run) -> List[TRECCovidExample]:
        queries = []
        query_xml_tree = ElementTree.parse(path)
        for topic in query_xml_tree.getroot():
            qid = topic.attrib["number"]
            query = topic.find("query").text
            queries.append(
                TRECCovidExample(
                    qid=qid,
                    text=query,
                    candidates=run[qid],
                    relevant_candidates=qrels[qid]))
        return queries

    @classmethod
    def from_folder(cls,
                    folder: str) -> 'TRECCovidDataset':
        query_path = os.path.join(folder, "topics.covid-round5.xml")
        qrels_path = os.path.join(folder, "qrels-covid_d5_j4.5-5.txt")
        run_path = os.path.join(folder, "expanded.anserini.final-r5.fusion1.txt")
        return cls(examples=cls.load_queries(query_path,
                                             cls.load_qrels(qrels_path),
                                             cls.load_run(run_path)))

    def query_document_tuples(self):
        return [((ex.qid, ex.text, ex.relevant_candidates), perm_pas)
                for ex in self.examples
                for perm_pas in permutations(ex.candidates, r=1)]

    def to_relevance_examples(self,
                              index_path: str) -> List[RelevanceExample]:
        loader = Cord19AbstractLoader(index_path)
        example_map = {}
        for (qid, text, rel_cands), cands in tqdm(self.query_document_tuples()):
            if qid not in example_map:
                example_map[qid] = [convert_to_unicode(text), [], [], [], []]
            example_map[qid][1].append([cand for cand in cands][0])
            try:
                passages = [loader.load_document(cand) for cand in cands]
                # Sometimes this abstract is empty.
                example_map[qid][2].append(
                    [convert_to_unicode(passage.abstract)
                     for passage in passages][0])
                example_map[qid][4].append(
                    [convert_to_unicode(passage.title)
                     for passage in passages][0])
            except ValueError as e:
                logging.error(e)
                logging.warning('Skipping passages')
                continue
            example_map[qid][3].append(cands[0] in rel_cands)
        mean_stats = defaultdict(list)

        for ex in self.examples:
            int_rels = np.array(list(map(int, example_map[ex.qid][3])))
            p = int(int_rels.sum())
            mean_stats['Random P@1'].append(np.mean(int_rels))
            n = len(ex.candidates) - p
            N = len(ex.candidates)
            if len(ex.candidates) <= 1000:
                mean_stats['Random R@1000'].append(1 if 1 in int_rels else 0)
            numer = np.array([sp.comb(n, i) / (N - i) for i in range(0, n + 1) if i != N]) * p
            if n == N:
                numer = np.append(numer, 0)
            denom = np.array([sp.comb(N, i) for i in range(0, n + 1)])
            rr = 1 / np.arange(1, n + 2)
            rmrr = np.sum(numer * rr / denom)
            mean_stats['Random MRR'].append(rmrr)
            rmrr10 = np.sum(numer[:10] * rr[:10] / denom[:10])
            mean_stats['Random MRR@10'].append(rmrr10)
            ex_index = len(ex.candidates)
            for rel_cand in ex.relevant_candidates:
                if rel_cand in ex.candidates:
                    ex_index = min(ex.candidates.index(rel_cand), ex_index)
            mean_stats['Existing MRR'].append(1 / (ex_index + 1)
                                              if ex_index < len(ex.candidates)
                                              else 0)
            mean_stats['Existing MRR@10'].append(1 / (ex_index + 1) if ex_index < 10 else 0)
        for k, v in mean_stats.items():
            logging.info(f'{k}: {np.mean(v)}')
        rel = [RelevanceExample(Query(text=query_text, id=qid),
                                list(map(lambda s: Text(s[1], dict(docid=s[0]), title=s[2]),
                                         zip(cands, cands_text, title))), rel_cands)
               for qid, (query_text, cands, cands_text, rel_cands, title) in example_map.items()]
        return rel



import setuptools


with open('README.md') as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='pygaggle',
    version='0.0.3.1',
    author='PyGaggle Gaggle',
    author_email='rpradeep@uwaterloo.ca',
    description='A gaggle of rerankers for text ranking and question answering.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/castorini/pygaggle',
    install_requires=requirements,
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import os
from collections import OrderedDict, defaultdict
from typing import List, Set, DefaultDict
import logging
from itertools import permutations

from pydantic import BaseModel
import scipy.special as sp
import numpy as np
from tqdm import tqdm

from .relevance import RelevanceExample, MsMarcoPassageLoader
from pygaggle.rerank.base import Query, Text
from pygaggle.data.unicode import convert_to_unicode


__all__ = ['MsMarcoExample', 'MsMarcoDataset']


# MsMarcoExample represents a query along with its ranked and re-ranked candidates.
class MsMarcoExample(BaseModel):
    qid: str
    text: str
    candidates: List[str]
    relevant_candidates: Set[str]


class MsMarcoDataset(BaseModel):
    examples: List[MsMarcoExample]

    # Load qrels from the provided path and return a dictionary mapping
    # qid -> set({doc_id, doc_id...}) for all doc_ids with relevance over 1.
    @classmethod
    def load_qrels(cls, path: str) -> DefaultDict[str, Set[str]]:
        qrels = defaultdict(set)
        with open(path) as f:
            for line in f:
                qid, _, doc_id, relevance = line.rstrip().split('\t')
                if int(relevance) >= 1:
                    qrels[qid].add(doc_id)
        return qrels

    # Load a run from the provided path.  The run file contains mappings from
    # a query id and a doc title to a rank.  load_run returns a dictionary
    # mapping query ids to lists of doc titles sorted by ascending rank.
    @classmethod
    def load_run(cls, path: str):
        '''Returns OrderedDict[str, List[str]]'''
        run = OrderedDict()
        with open(path) as f:
            for i, line in enumerate(f):
                qid, doc_title, rank = line.split('\t')
                if qid not in run:
                    run[qid] = []
                run[qid].append((doc_title, int(rank)))
        sorted_run = OrderedDict()
        for qid, doc_titles_ranks in run.items():
            sorted(doc_titles_ranks, key=lambda x: x[1])
            doc_titles = [doc_titles for doc_titles, _ in doc_titles_ranks]
            sorted_run[qid] = doc_titles
        return sorted_run

    @classmethod
    def load_queries(cls,
                     path: str,
                     qrels: DefaultDict[str, Set[str]],
                     run) -> List[MsMarcoExample]:
        queries = []
        with open(path) as f:
            for line in f:
                qid, query = line.rstrip().split('\t')
                queries.append(MsMarcoExample(qid=qid,
                                              text=query,
                                              candidates=run[qid],
                                              relevant_candidates=qrels[qid]))
        return queries

    @classmethod
    def from_folder(cls,
                    folder: str,
                    split: str = 'dev',
                    is_duo: bool = False) -> 'MsMarcoDataset':
        run_mono = "mono." if is_duo else ""
        query_path = os.path.join(folder, f"queries.{split}.small.tsv")
        qrels_path = os.path.join(folder, f"qrels.{split}.small.tsv")
        run_path = os.path.join(folder, f"run.{run_mono}{split}.small.tsv")
        return cls(examples=cls.load_queries(query_path,
                                             cls.load_qrels(qrels_path),
                                             cls.load_run(run_path)))

    def query_passage_tuples(self, is_duo: bool = False):
        return [((ex.qid, ex.text, ex.relevant_candidates), perm_pas)
                for ex in self.examples
                for perm_pas in permutations(ex.candidates, r=1+int(is_duo))]

    def to_relevance_examples(self,
                              index_path: str,
                              is_duo: bool = False) -> List[RelevanceExample]:
        loader = MsMarcoPassageLoader(index_path)
        example_map = {}
        for (qid, text, rel_cands), cands in tqdm(self.query_passage_tuples()):
            if qid not in example_map:
                example_map[qid] = [convert_to_unicode(text), [], [], []]
            example_map[qid][1].append([cand for cand in cands][0])
            try:
                passages = [loader.load_passage(cand) for cand in cands]
                example_map[qid][2].append(
                    [convert_to_unicode(passage.all_text)
                     for passage in passages][0])
            except ValueError:
                logging.warning(f'Skipping {passages}')
                continue
            example_map[qid][3].append(cands[0] in rel_cands)
        mean_stats = defaultdict(list)
        for ex in self.examples:
            int_rels = np.array(list(map(int, example_map[ex.qid][3])))
            p = int_rels.sum()/(len(ex.candidates) - 1) if is_duo else int_rels.sum()
            mean_stats['Random P@1'].append(np.mean(int_rels))
            n = len(ex.candidates) - p
            N = len(ex.candidates)
            if len(ex.candidates) <= 1000:
                mean_stats['Random R@1000'].append(1 if 1 in int_rels else 0)
            numer = np.array([sp.comb(n, i) / (N - i) for i in range(0, n + 1) if i != N]) * p
            if n == N:
                numer = np.append(numer, 0)
            denom = np.array([sp.comb(N, i) for i in range(0, n + 1)])
            rr = 1 / np.arange(1, n + 2)
            rmrr = np.sum(numer * rr / denom)
            mean_stats['Random MRR'].append(rmrr)
            rmrr10 = np.sum(numer[:10] * rr[:10] / denom[:10])
            mean_stats['Random MRR@10'].append(rmrr10)
            ex_index = len(ex.candidates)
            for rel_cand in ex.relevant_candidates:
                if rel_cand in ex.candidates:
                    ex_index = min(ex.candidates.index(rel_cand), ex_index)
            mean_stats['Existing MRR'].append(1 / (ex_index + 1)
                                              if ex_index < len(ex.candidates)
                                              else 0)
            mean_stats['Existing MRR@10'].append(1 / (ex_index + 1) if ex_index < 10 else 0)
        for k, v in mean_stats.items():
            logging.info(f'{k}: {np.mean(v)}')
        return [RelevanceExample(Query(text=query_text, id=qid),
                                 list(map(lambda s: Text(s[1], dict(docid=s[0])),
                                          zip(cands, cands_text))),
                                 rel_cands)
                for qid, (query_text, cands, cands_text, rel_cands) in example_map.items()]



from typing import Optional, List
from pathlib import Path
import logging

from pydantic import BaseModel, validator
from transformers import (AutoModel,
                          AutoTokenizer)
import torch

from .args import ArgumentParserBuilder, opt
from pygaggle.rerank.base import Reranker
from pygaggle.rerank.bm25 import Bm25Reranker
from pygaggle.rerank.transformer import (
    UnsupervisedTransformerReranker,
    MonoT5,
    MonoBERT
)
from pygaggle.rerank.random import RandomReranker
from pygaggle.rerank.similarity import CosineSimilarityMatrixProvider
from pygaggle.model import (SimpleBatchTokenizer,
                            RerankerEvaluator,
                            metric_names,
                            MsMarcoWriter)
from pygaggle.data import MsMarcoDataset
from pygaggle.settings import MsMarcoSettings


SETTINGS = MsMarcoSettings()
METHOD_CHOICES = ('transformer', 'bm25', 't5', 'seq_class_transformer',
                  'random')


class DocumentRankingEvaluationOptions(BaseModel):
    task: str
    dataset: Path
    index_dir: Path
    method: str
    model: str
    split: str
    batch_size: int
    seg_size: int
    seg_stride: int
    aggregate_method: str
    device: str
    is_duo: bool
    from_tf: bool
    metrics: List[str]
    model_type: Optional[str]
    tokenizer_name: Optional[str]

    @validator('task')
    def task_exists(cls, v: str):
        assert v in ['msmarco']

    @validator('dataset')
    def dataset_exists(cls, v: Path):
        assert v.exists(), 'data directory must exist'
        return v

    @validator('index_dir')
    def index_dir_exists(cls, v: Path):
        assert v.exists(), 'index directory must exist'
        return v

    @validator('model')
    def model_sane(cls, v: str, values, **kwargs):
        method = values['method']
        if method == 'transformer' and v is None:
            raise ValueError('transformer name or path must be specified')
        return v

    @validator('tokenizer_name')
    def tokenizer_sane(cls, v: str, values, **kwargs):
        if v is None:
            return values['model']
        return v


def construct_t5(options: DocumentRankingEvaluationOptions) -> Reranker:
    model = MonoT5.get_model(options.model,
                             from_tf=options.from_tf,
                             device=options.device)
    tokenizer = MonoT5.get_tokenizer(options.model_type, batch_size=options.batch_size)
    return MonoT5(model, tokenizer)


def construct_transformer(options:
                          DocumentRankingEvaluationOptions) -> Reranker:
    device = torch.device(options.device)
    model = AutoModel.from_pretrained(options.model,
                                      from_tf=options.from_tf).to(device).eval()
    tokenizer = SimpleBatchTokenizer(AutoTokenizer.from_pretrained(
        options.tokenizer_name, use_fast=False,),
        options.batch_size)
    provider = CosineSimilarityMatrixProvider()
    return UnsupervisedTransformerReranker(model, tokenizer, provider)


def construct_seq_class_transformer(options: DocumentRankingEvaluationOptions
                                    ) -> Reranker:
    model = MonoBERT.get_model(options.model, from_tf=options.from_tf, device=options.device)
    tokenizer = MonoBERT.get_tokenizer(options.tokenizer_name)
    return MonoBERT(model, tokenizer)


def construct_bm25(options: DocumentRankingEvaluationOptions) -> Reranker:
    return Bm25Reranker(index_path=str(options.index_dir))


def main():
    apb = ArgumentParserBuilder()
    apb.add_opts(opt('--task',
                     type=str,
                     default='msmarco'),
                 opt('--dataset', type=Path, required=True),
                 opt('--index-dir', type=Path, required=True),
                 opt('--method',
                     required=True,
                     type=str,
                     choices=METHOD_CHOICES),
                 opt('--model',
                     required=True,
                     type=str,
                     help='Path to pre-trained model or huggingface model name'),
                 opt('--output-file', type=Path, default='.'),
                 opt('--overwrite-output', action='store_true'),
                 opt('--split',
                     type=str,
                     default='dev',
                     choices=('dev', 'eval')),
                 opt('--batch-size', '-bsz', type=int, default=96),
                 opt('--device', type=str, default='cuda:0'),
                 opt('--is-duo', action='store_true'),
                 opt('--from-tf', action='store_true'),
                 opt('--metrics',
                     type=str,
                     nargs='+',
                     default=metric_names(),
                     choices=metric_names()),
                 opt('--model-type', type=str),
                 opt('--tokenizer-name', type=str),
                 opt('--seg-size', type=int, default=10),
                 opt('--seg-stride', type=int, default=5),
                 opt('--aggregate-method', type=str, default="max"))
    args = apb.parser.parse_args()
    options = DocumentRankingEvaluationOptions(**vars(args))
    logging.info("Preprocessing Queries & Docs:")
    ds = MsMarcoDataset.from_folder(str(options.dataset), split=options.split,
                                    is_duo=options.is_duo)
    examples = ds.to_relevance_examples(str(options.index_dir),
                                        is_duo=options.is_duo)
    logging.info("Loading Ranker & Tokenizer:")
    construct_map = dict(transformer=construct_transformer,
                         bm25=construct_bm25,
                         t5=construct_t5,
                         seq_class_transformer=construct_seq_class_transformer,
                         random=lambda _: RandomReranker())
    reranker = construct_map[options.method](options)
    writer = MsMarcoWriter(args.output_file, args.overwrite_output)
    evaluator = RerankerEvaluator(reranker, options.metrics, writer=writer)
    width = max(map(len, args.metrics)) + 1
    logging.info("Reranking:")
    for metric in evaluator.evaluate_by_segments(examples,
                                                 options.seg_size,
                                                 options.seg_stride,
                                                 options.aggregate_method):
        logging.info(f'{metric.name:<{width}}{metric.value:.5}')


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import List, Optional
import abc

from pygaggle.data.relevance import RelevanceExample

__all__ = ['Writer', 'MsMarcoWriter', 'TrecWriter']


class Writer:
    def __init__(self, path: Optional[Path] = None, overwrite: bool = True, tag: Optional[str] = None):
        self.to_output = str(path) not in [".", None]
        print(f'Writing run: {self.to_output}')
        if self.to_output:
            self.f = open(path, "w" if overwrite else "w+")
        self.tag = tag

    def write_line(self, text: str):
        if self.to_output:
            self.f.write(f"{text}\n")

    @abc.abstractmethod
    def write(self, scores: List[float], example: RelevanceExample):
        pass


class MsMarcoWriter(Writer):
    def write(self, scores: List[float], example: RelevanceExample):
        doc_scores = sorted(list(zip(example.documents, scores)),
                            key=lambda x: x[1], reverse=True)
        for ct, (doc, score) in enumerate(doc_scores):
            self.write_line(f"{example.query.id}\t{doc.metadata['docid']}\t{ct+1}")


class TrecWriter(Writer):
    def write(self, scores: List[float], example: RelevanceExample):
        doc_scores = sorted(list(zip(example.documents, scores)),
                            key=lambda x: x[1], reverse=True)
        for ct, (doc, score) in enumerate(doc_scores):
            self.write_line(f"{example.query.id}\tQ0\t{doc.metadata['docid']}\t{ct+1}\t{score}\t{self.tag}")



import coloredlogs
import logging


__all__ = []


coloredlogs.install(level='INFO',
                    fmt='%(asctime)s [%(levelname)s] %(module)s: %(message)s')
logging.getLogger("transformers").setLevel(logging.WARNING)
logging.getLogger("filelock").setLevel(logging.WARNING)



import argparse


__all__ = ['ArgumentParserBuilder', 'opt']


def _make_parser_setter(option, key):
    def fn(value):
        option.kwargs[key] = value
        return option
    return fn


class ArgumentParserOption:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        return iter((self.args, self.kwargs))

    def __getattr__(self, item: str):
        if item == 'kwargs':
            return self.kwargs
        if item == 'args':
            return self.args
        return _make_parser_setter(self, item)


opt = ArgumentParserOption


class ArgumentParserBuilder(object):
    def __init__(self, **init_kwargs):
        self.parser = argparse.ArgumentParser(**init_kwargs)

    def add_opts(self, *options):
        for args, kwargs in options:
            self.parser.add_argument(*args, **kwargs)
        return self.parser
